# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Integration tests for fast-loading FF and PP files."""

# import iris tests first so that some things can be initialised
# before importing anything else.
import iris.tests as tests

from collections.abc import Iterable
import tempfile
import shutil

import numpy as np

import iris.coords
from iris.coords import DimCoord, AuxCoord, CellMethod
from iris.coord_systems import GeogCS
from iris.cube import Cube, CubeList
from iris.exceptions import IgnoreCubeException
from iris.fileformats.pp import EARTH_RADIUS, STASH
from iris.fileformats.um._fast_load import STRUCTURED_LOAD_CONTROLS

import iris


class Mixin_FieldTest:
    # A mixin providing common facilities for fast-load testing :
    #   * create 'raw' cubes to produce the desired PP fields in a test file.
    #   * save 'raw' cubes to temporary PP files that get deleted afterwards.
    #   * control whether tests run with 'normal' or 'fast' loading.

    def setUp(self):
        # Create a private temporary directory.
        self.temp_dir_path = tempfile.mkdtemp()
        # Initialise temporary filename generation.
        self.tempfile_count = 0
        self.tempfile_path_fmt = (
            "{dir_path}/tempfile_{prefix}_{file_number:06d}{suffix}"
        )
        # Enable fast loading, if the inheritor enables it.
        # N.B. *requires* the user to define "self.do_fast_loads" (no default).
        if self.do_fast_loads:
            # Enter a 'structured load' context.
            self.load_context = STRUCTURED_LOAD_CONTROLS.context(
                loads_use_structured=True
            )
            # N.B. we can't use a 'with', so issue separate 'enter' and 'exit'
            # calls instead.
            self.load_context.__enter__()

    def tearDown(self):
        # Delete temporary directory.
        shutil.rmtree(self.temp_dir_path)
        if self.do_fast_loads:
            # End the 'fast loading' context.
            self.load_context.__exit__(None, None, None)

    def _temp_filepath(self, user_name="", suffix=".pp"):
        # Return the filepath for a new temporary file.
        self.tempfile_count += 1
        file_path = self.tempfile_path_fmt.format(
            dir_path=self.temp_dir_path,
            prefix=user_name,
            file_number=self.tempfile_count,
            suffix=suffix,
        )
        return file_path

    def save_fieldcubes(self, cubes, basename=""):
        # Save cubes to a temporary file, and return its filepath.
        file_path = self._temp_filepath(user_name=basename, suffix=".pp")
        iris.save(cubes, file_path)
        return file_path

    def fields(
        self,
        c_t=None,
        cft=None,
        ctp=None,
        c_h=None,
        c_p=None,
        phn=0,
        mmm=None,
        pse=None,
    ):
        # Return a list of 2d cubes representing raw PPFields, from args
        # specifying sequences of (scalar) coordinate values.
        # TODO? : add bounds somehow ?
        #
        # Arguments 'c<xx>' are either a single int value, making a scalar
        # coord, or a string of characters : '0'-'9' (index) or '-' (missing).
        # The indexes select point values from fixed list of possibles.
        #
        # Argument 'c_h' and 'c_p' represent height or pressure values, so
        # ought to be mutually exclusive -- these control LBVC.
        #
        # Argument 'phn' indexes phenomenon types.
        #
        # Argument 'mmm' denotes existence (or not) of a cell method of type
        # 'average' or 'min' or 'max' (values '012' respectively), applying to
        # the time values -- ultimately, this controls LBTIM.
        #
        # Argument 'pse' denotes pseudo-level numbers.
        # These translate into 'LBUSER5' values.

        # Get the number of result cubes, defined by the 'longest' arg.
        def arglen(arg):
            # Get the 'length' of a control argument.
            if arg is None:
                result = 0
            elif isinstance(arg, str):
                result = len(arg)
            else:
                result = 1
            return result

        n_flds = max(arglen(x) for x in (c_t, cft, ctp, c_h, c_p, mmm))

        # Make basic anonymous test cubes.
        ny, nx = 3, 5
        data = np.arange(n_flds * ny * nx, dtype=np.float32)
        data = data.reshape((n_flds, ny, nx))
        cubes = [Cube(data[i]) for i in range(n_flds)]

        # Define test point values for making coordinates.
        time_unit = "hours since 1970-01-01"
        period_unit = "hours"
        height_unit = "m"
        pressure_unit = "hPa"
        time_values = 24.0 * np.arange(10)
        height_values = 100.0 * np.arange(1, 11)
        pressure_values = [
            100.0,
            150.0,
            200.0,
            250.0,
            300.0,
            500.0,
            850.0,
            1000.0,
        ]
        pseudolevel_values = range(1, 11)  # A valid value is >= 1.

        # Test phenomenon details.
        # NOTE: in order to write/readback as identical, these also contain a
        # canonical unit and matching STASH attribute.
        # Those could in principle be looked up, but it's a bit awkward.
        phenomenon_values = [
            ("air_temperature", "K", "m01s01i004"),
            ("x_wind", "m s-1", "m01s00i002"),
            ("y_wind", "m s-1", "m01s00i003"),
            ("specific_humidity", "kg kg-1", "m01s00i010"),
        ]

        # Test cell-methods.
        # NOTE: if you add an *interval* to any of these cell-methods, it is
        # not saved into the PP file (?? or maybe not loaded back again ??).
        # This could be a PP save/load bug, or maybe just because no bounds ?
        cell_method_values = [
            CellMethod("mean", "time"),
            CellMethod("maximum", "time"),
            CellMethod("minimum", "time"),
        ]

        # Define helper to decode an argument as a list of test values.
        def arg_vals(arg, vals):
            # Decode an argument to a list of 'n_flds' coordinate point values.
            # (or 'None' where missing)

            # First get a list of value indices from the argument.
            # Can be: a single index value; a list of indices; or a string.
            if isinstance(arg, Iterable) and not isinstance(arg, str):
                # Can also just pass a simple iterable of values.
                inds = [int(val) for val in arg]
            else:
                n_vals = arglen(arg)
                if n_vals == 0:
                    inds = [None] * n_flds
                elif n_vals == 1:
                    inds = [int(arg)] * n_flds
                else:
                    assert isinstance(arg, str)
                    inds = [None if char == "-" else int(char) for char in arg]

            # Convert indices to selected point values.
            values = [None if ind is None else vals[int(ind)] for ind in inds]

            return values

        # Apply phenomenon_values definitions.
        phenomena = arg_vals(phn, phenomenon_values)
        for cube, (name, units, stash) in zip(cubes, phenomena):
            cube.rename(name)
            # NOTE: in order to get a cube that will write+readback the same,
            # the units must be the canonical one.
            cube.units = units
            # NOTE: in order to get a cube that will write+readback the same,
            # we must include a STASH attribute.
            cube.attributes["STASH"] = STASH.from_msi(stash)
            cube.fill_value = np.float32(-1e30)

        # Add x and y coords.
        cs = GeogCS(EARTH_RADIUS)
        xvals = np.linspace(0.0, 180.0, nx)
        co_x = DimCoord(
            np.array(xvals, dtype=np.float32),
            standard_name="longitude",
            units="degrees",
            coord_system=cs,
        )
        yvals = np.linspace(-45.0, 45.0, ny)
        co_y = DimCoord(
            np.array(yvals, dtype=np.float32),
            standard_name="latitude",
            units="degrees",
            coord_system=cs,
        )
        for cube in cubes:
            cube.add_dim_coord(co_y, 0)
            cube.add_dim_coord(co_x, 1)

        # Add multiple scalar coordinates as defined by the arguments.
        def arg_coords(arg, name, unit, vals=None):
            # Decode an argument to a list of scalar coordinates.
            if vals is None:
                vals = np.arange(n_flds + 2)  # Note allowance
            vals = arg_vals(arg, vals)
            coords = [
                None if val is None else DimCoord([val], units=unit)
                for val in vals
            ]
            # Apply names separately, as 'pressure' is not a standard name.
            for coord in coords:
                if coord:
                    coord.rename(name)
                    # Also fix heights to match what comes from a PP file.
                    if name == "height":
                        coord.attributes["positive"] = "up"
            return coords

        def add_arg_coords(arg, name, unit, vals=None):
            # Add scalar coordinates to each cube, for one argument.
            coords = arg_coords(arg, name, unit, vals)
            for cube, coord in zip(cubes, coords):
                if coord:
                    cube.add_aux_coord(coord)

        add_arg_coords(c_t, "time", time_unit, time_values)
        add_arg_coords(cft, "forecast_reference_time", time_unit)
        add_arg_coords(ctp, "forecast_period", period_unit, time_values)
        add_arg_coords(c_h, "height", height_unit, height_values)
        add_arg_coords(c_p, "pressure", pressure_unit, pressure_values)
        add_arg_coords(pse, "pseudo_level", "1", pseudolevel_values)

        # Add cell methods as required.
        methods = arg_vals(mmm, cell_method_values)
        for cube, method in zip(cubes, methods):
            if method:
                cube.add_cell_method(method)

        return cubes


class MixinBasic:
    # A mixin of tests that can be applied to *either* standard or fast load.
    # The "real" test classes must inherit this, and Mixin_FieldTest,
    # and define 'self.do_fast_loads' as True or False.
    #
    # Basic functional tests.

    def test_basic(self):
        # Show that basic load merging works.
        flds = self.fields(c_t="123", cft="000", ctp="123", c_p=0)
        file = self.save_fieldcubes(flds)
        results = iris.load(file)
        expected = CubeList(flds).merge()
        self.assertEqual(results, expected)

    def test_phenomena(self):
        # Show that different phenomena are merged into distinct cubes.
        flds = self.fields(c_t="1122", phn="0101")
        file = self.save_fieldcubes(flds)
        results = iris.load(file)
        expected = CubeList(flds).merge()
        self.assertEqual(results, expected)

    def test_cross_file_concatenate(self):
        # Combine vector dimensions (i.e. concatenate) across multiple files.
        fldset_1 = self.fields(c_t="12")
        fldset_2 = self.fields(c_t="34")
        file_1 = self.save_fieldcubes(fldset_1)
        file_2 = self.save_fieldcubes(fldset_2)
        results = iris.load((file_1, file_2))
        expected = CubeList(fldset_1 + fldset_2).merge()
        self.assertEqual(results, expected)

    def test_cell_method(self):
        # Check that cell methods (i.e. LBPROC values) produce distinct
        # phenomena.
        flds = self.fields(c_t="000111222", mmm="-01-01-01")
        file = self.save_fieldcubes(flds)
        results = iris.load(file)
        expected = CubeList(
            CubeList(flds[i_start::3]).merge_cube() for i_start in range(3)
        )
        self.assertEqual(results, expected)


class MixinCallDetails:
    # A mixin of tests that can be applied to *either* standard or fast load.
    # The "real" test classes must inherit this, and Mixin_FieldTest,
    # and define 'self.do_fast_loads' as True or False.
    #
    # Tests for different load calls and load-call arguments.

    def test_stash_constraint(self):
        # Check that an attribute constraint functions correctly.
        # Note: this is a special case in "fileformats.pp".
        flds = self.fields(c_t="1122", phn="0101")
        file = self.save_fieldcubes(flds)
        airtemp_flds = [fld for fld in flds if fld.name() == "air_temperature"]
        stash_attribute = airtemp_flds[0].attributes["STASH"]
        results = iris.load(
            file, iris.AttributeConstraint(STASH=stash_attribute)
        )
        expected = CubeList(airtemp_flds).merge()
        self.assertEqual(results, expected)

    def test_ordinary_constraint(self):
        # Check that a 'normal' constraint functions correctly.
        # Note: *should* be independent of structured loading.
        flds = self.fields(c_h="0123")
        file = self.save_fieldcubes(flds)
        height_constraint = iris.Constraint(height=lambda h: 150.0 < h < 350.0)
        results = iris.load(file, height_constraint)
        expected = CubeList(flds[1:3]).merge()
        self.assertEqual(results, expected)

    def test_callback(self):
        # Use 2 timesteps each of (air-temp on height) and (rh on pressure).
        flds = self.fields(c_t="0011", phn="0303", c_h="0-1-", c_p="-2-3")
        file = self.save_fieldcubes(flds)

        if not self.do_fast_loads:

            def callback(cube, field, filename):
                self.assertEqual(filename, file)
                lbvc = field.lbvc
                if lbvc == 1:
                    # reject the height level data (accept only pressure).
                    raise IgnoreCubeException()
                else:
                    # Record the LBVC value.
                    cube.attributes["LBVC"] = lbvc

        else:

            def callback(cube, collation, filename):
                self.assertEqual(filename, file)
                lbvcs = [fld.lbvc for fld in collation.fields]
                lbvc0 = lbvcs[0]
                if not np.all(lbvcs == lbvc0):
                    msg = "Fields have different LBVCs : {}"
                    raise ValueError(msg.format(set(lbvcs)))
                if lbvc0 == 1:
                    # reject the height level data (accept only pressure).
                    raise IgnoreCubeException()
                else:
                    # Record the LBVC values.
                    cube.attributes["A_LBVC"] = lbvcs

        results = iris.load(file, callback=callback)

        # Make an 'expected' from selected fields, with the expected attribute.
        expected = CubeList([flds[1], flds[3]]).merge()
        if not self.do_fast_loads:
            # This is actually a NumPy int32, so honour that here.
            expected[0].attributes["LBVC"] = np.int32(8)
        else:
            expected[0].attributes["A_LBVC"] = [8, 8]

        self.assertEqual(results, expected)

    def test_load_cube(self):
        flds = self.fields(c_t="123", cft="000", ctp="123", c_p=0)
        file = self.save_fieldcubes(flds)
        results = iris.load_cube(file)
        expected = CubeList(flds).merge_cube()
        self.assertEqual(results, expected)

    def test_load_cubes(self):
        flds = self.fields(c_h="0123")
        file = self.save_fieldcubes(flds)
        height_constraints = [
            iris.Constraint(height=300.0),
            iris.Constraint(height=lambda h: 150.0 < h < 350.0),
            iris.Constraint("air_temperature"),
        ]
        results = iris.load_cubes(file, height_constraints)
        expected = CubeList(
            [
                flds[2],
                CubeList(flds[1:3]).merge_cube(),
                CubeList(flds).merge_cube(),
            ]
        )
        self.assertEqual(results, expected)

    def test_load_raw(self):
        fldset_1 = self.fields(c_t="015", phn="001")
        fldset_2 = self.fields(c_t="234")
        file_1 = self.save_fieldcubes(fldset_1)
        file_2 = self.save_fieldcubes(fldset_2)
        results = iris.load_raw((file_1, file_2))
        if not self.do_fast_loads:
            # Each 'raw' cube is just one field.
            expected = CubeList(fldset_1 + fldset_2)
        else:
            # 'Raw' cubes have combined (vector) times within each file.
            # The 'other' phenomenon appears seperately.
            expected = CubeList(
                [
                    CubeList(fldset_1[:2]).merge_cube(),
                    CubeList(fldset_2).merge_cube(),
                    fldset_1[2],
                ]
            )

        # Again here, the order of these results is not stable :
        # It varies with random characters in the temporary filepath.
        #
        # *****************************************************************
        # *** Here, this is clearly ALSO the case for "standard" loads. ***
        # *****************************************************************
        #
        # E.G. run "test_fast_load.py -v TestCallDetails__Iris.test_load_raw" :
        # If you remove the sort operations, this fails "sometimes".
        #
        # To fix this, sort both expected and results by (first) timepoint
        # - for which purpose we made all the time values different.

        def timeorder(cube):
            return cube.coord("time").points[0]

        expected = sorted(expected, key=timeorder)
        results = sorted(results, key=timeorder)

        self.assertEqual(results, expected)


class MixinDimsAndOrdering:
    # A mixin of tests that can be applied to *either* standard or fast load.
    # The "real" test classes must inherit this, and Mixin_FieldTest,
    # and define 'self.do_fast_loads' as True or False.
    #
    # Tests for multidimensional results and dimension orderings.

    def test_multidim(self):
        # Check that a full 2-phenom * 2d structure all works properly.
        flds = self.fields(c_t="00001111", c_h="00110011", phn="01010101")
        file = self.save_fieldcubes(flds)
        results = iris.load(file)
        expected = CubeList(flds).merge()
        self.assertEqual(results, expected)

    def test_odd_order(self):
        # Show that an erratic interleaving of phenomena fields still works.
        # N.B. field sequences *within* each phenomenon are properly ordered.
        flds = self.fields(c_t="00010111", c_h="00101101", phn="01001011")
        file = self.save_fieldcubes(flds)
        results = iris.load(file)
        expected = CubeList(flds).merge()
        self.assertEqual(results, expected)

    def test_v_t_order(self):
        # With height varying faster than time, first dimension is time,
        # which matches the 'normal' load behaviour.
        flds = self.fields(c_t="000111", c_h="012012")
        file = self.save_fieldcubes(flds)
        results = iris.load(file)
        expected = CubeList(flds).merge()
        # Order is (t, h, y, x), which is "standard".
        self.assertEqual(expected[0].coord_dims("time"), (0,))
        self.assertEqual(expected[0].coord_dims("height"), (1,))
        self.assertEqual(results, expected)

    def test_t_v_order(self):
        # With time varying faster than height, first dimension is height,
        # which does not match the 'normal' load.
        flds = self.fields(c_t="010101", c_h="001122")
        file = self.save_fieldcubes(flds)
        results = iris.load(file)
        expected = CubeList(flds).merge()
        if not self.do_fast_loads:
            # Order is (t, h, y, x), which is "standard".
            self.assertEqual(results[0].coord_dims("time"), (0,))
            self.assertEqual(results[0].coord_dims("height"), (1,))
        else:
            # Order is (h, t, y, x), which is *not* "standard".
            self.assertEqual(results[0].coord_dims("time"), (1,))
            self.assertEqual(results[0].coord_dims("height"), (0,))
            expected[0].transpose((1, 0, 2, 3))
        self.assertEqual(results, expected)

    def test_missing_combination(self):
        # A case where one field is 'missing' to make a 2d result.
        flds = self.fields(c_t="00011", c_h="01202")
        file = self.save_fieldcubes(flds)
        results = iris.load(file)
        expected = CubeList(flds).merge()
        self.assertEqual(expected[0].coord_dims("time"), (0,))
        self.assertEqual(expected[0].coord_dims("height"), (0,))
        if self.do_fast_loads:
            # Something a bit weird happens to the 'height' coordinate in this
            # case (and not for standard load).
            for cube in expected:
                cube.coord("height").points = np.array(
                    cube.coord("height").points, dtype=np.float32
                )
                cube.coord("height").attributes = {}
        self.assertEqual(results, expected)


class MixinProblemCases:
    def test_FAIL_scalar_vector_concatenate(self):
        # Structured load can produce a scalar coordinate from one file, and a
        # matching vector one from another file, but these won't "combine".
        # We'd really like to fix this one...
        (single_timepoint_fld,) = self.fields(c_t="1")
        multi_timepoint_flds = self.fields(c_t="23")
        file_single = self.save_fieldcubes(
            [single_timepoint_fld], basename="single"
        )
        file_multi = self.save_fieldcubes(
            multi_timepoint_flds, basename="multi"
        )

        results = iris.load((file_single, file_multi))
        if not self.do_fast_loads:
            # This is what we'd LIKE to get (what iris.load gives).
            expected = CubeList(
                multi_timepoint_flds + [single_timepoint_fld]
            ).merge()
        else:
            # This is what we ACTUALLY get at present.
            # It can't combine the scalar and vector time coords.
            expected = CubeList(
                [
                    CubeList(multi_timepoint_flds).merge_cube(),
                    single_timepoint_fld,
                ]
            )
            # NOTE: in this case, we need to sort the results to ensure a
            # repeatable ordering, because ??somehow?? the random temporary
            # directory name affects the ordering of the cubes in the result !
            results = CubeList(sorted(results, key=lambda cube: cube.shape))
        self.assertEqual(results, expected)

    def test_FAIL_phenomena_nostash(self):
        # If we remove the 'STASH' attributes, certain phenomena can still be
        # successfully encoded+decoded by standard load using LBFC values.
        # Structured loading gets this wrong, because it does not use LBFC in
        # characterising phenomena.
        flds = self.fields(c_t="1122", phn="0101")
        for fld in flds:
            del fld.attributes["STASH"]
        file = self.save_fieldcubes(flds)
        results = iris.load(file)
        if not self.do_fast_loads:
            # This is what we'd LIKE to get (what iris.load gives).
            expected = CubeList(flds).merge()
        else:
            # At present, we get a cube incorrectly combined together over all
            # 4 timepoints, with the same phenomenon for all (!wrong!).
            # It's a bit tricky to arrange the existing data like that.
            # Do it by hacking the time values to allow merge, and then fixing
            # up the time
            old_t1, old_t2 = (
                fld.coord("time").points[0] for fld in (flds[0], flds[2])
            )
            for i_fld, fld in enumerate(flds):
                # Hack the phenomena to all look like the first one.
                fld.rename("air_temperature")
                fld.units = "K"
                # Hack the time points so the 4 cube can merge into one.
                fld.coord("time").points = [old_t1 + i_fld]
            one_cube = CubeList(flds).merge_cube()
            # Replace time dim with an anonymous dim.
            co_t_fake = one_cube.coord("time")
            one_cube.remove_coord(co_t_fake)
            # Reconstruct + add back the expected auxiliary time coord.
            co_t_new = AuxCoord(
                [old_t1, old_t1, old_t2, old_t2],
                standard_name="time",
                units=co_t_fake.units,
            )
            one_cube.add_aux_coord(co_t_new, 0)
            expected = [one_cube]
        self.assertEqual(results, expected)

    def test_FAIL_pseudo_levels(self):
        # Show how pseudo levels are handled.
        flds = self.fields(c_t="000111222", pse="123123123")
        file = self.save_fieldcubes(flds)
        results = iris.load(file)
        expected = CubeList(flds).merge()

        # NOTE: this problem is now fixed : Structured load gives the same answer.
        #
        #        if not self.do_fast_loads:
        #            expected = CubeList(flds).merge()
        #        else:
        #            # Structured loading doesn't understand pseudo-level.
        #            # The result is rather horrible...
        #
        #            # First get a cube over 9 timepoints.
        #            flds = self.fields(c_t='012345678',
        #                               pse=1)  # result gets level==2, not clear why.
        #
        #            # Replace the time coord with an AUX coord.
        #            nine_timepoints_cube = CubeList(flds).merge_cube()
        #            co_time = nine_timepoints_cube.coord('time')
        #            nine_timepoints_cube.remove_coord(co_time)
        #            nine_timepoints_cube.add_aux_coord(AuxCoord.from_coord(co_time),
        #                                               0)
        #            # Set the expected timepoints equivalent to '000111222'.
        #            nine_timepoints_cube.coord('time').points = \
        #                np.array([0.0, 0.0, 0.0, 24.0, 24.0, 24.0, 48.0, 48.0, 48.0])
        #            # Make a cubelist with this single cube.
        #            expected = CubeList([nine_timepoints_cube])

        self.assertEqual(results, expected)


class TestBasic__Iris(Mixin_FieldTest, MixinBasic, tests.IrisTest):
    # Finally, an actual test-class (unittest.TestCase) :
    # run the 'basic' tests with *normal* loading.
    do_fast_loads = False


class TestBasic__Fast(Mixin_FieldTest, MixinBasic, tests.IrisTest):
    # Finally, an actual test-class (unittest.TestCase) :
    # run the 'basic' tests with *FAST* loading.
    do_fast_loads = True


class TestCallDetails__Iris(Mixin_FieldTest, MixinCallDetails, tests.IrisTest):
    # Finally, an actual test-class (unittest.TestCase) :
    # run the 'call details' tests with *normal* loading.
    do_fast_loads = False


class TestCallDetails__Fast(Mixin_FieldTest, MixinCallDetails, tests.IrisTest):
    # Finally, an actual test-class (unittest.TestCase) :
    # run the 'call details' tests with *FAST* loading.
    do_fast_loads = True


class TestDimsAndOrdering__Iris(
    Mixin_FieldTest, MixinDimsAndOrdering, tests.IrisTest
):
    # Finally, an actual test-class (unittest.TestCase) :
    # run the 'dimensions and ordering' tests with *normal* loading.
    do_fast_loads = False


class TestDimsAndOrdering__Fast(
    Mixin_FieldTest, MixinDimsAndOrdering, tests.IrisTest
):
    # Finally, an actual test-class (unittest.TestCase) :
    # run the 'dimensions and ordering' tests with *FAST* loading.
    do_fast_loads = True


class TestProblems__Iris(Mixin_FieldTest, MixinProblemCases, tests.IrisTest):
    # Finally, an actual test-class (unittest.TestCase) :
    # run the 'failure cases' tests with *normal* loading.
    do_fast_loads = False


class TestProblems__Fast(Mixin_FieldTest, MixinProblemCases, tests.IrisTest):
    # Finally, an actual test-class (unittest.TestCase) :
    # run the 'failure cases' tests with *FAST* loading.
    do_fast_loads = True


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Integration tests for loading LBC fieldsfiles."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from unittest import mock

import numpy as np

import iris


@tests.skip_data
class TestLBC(tests.IrisTest):
    def setUp(self):
        # Load multiple cubes from a test file.
        file_path = tests.get_data_path(("FF", "lbc", "small_lbc"))
        self.all_cubes = iris.load(file_path)
        # Select the second cube for detailed checks (the first is orography).
        self.test_cube = self.all_cubes[1]

    def test_various_cubes_shapes(self):
        # Check a few aspects of the loaded cubes.
        cubes = self.all_cubes
        self.assertEqual(len(cubes), 10)
        self.assertEqual(cubes[0].shape, (16, 16))
        self.assertEqual(cubes[1].shape, (2, 4, 16, 16))
        self.assertEqual(cubes[3].shape, (2, 5, 16, 16))

    def test_cube_coords(self):
        # Check coordinates of one cube.
        cube = self.test_cube
        self.assertEqual(len(cube.coords()), 8)
        for name, shape in [
            ("forecast_reference_time", (1,)),
            ("time", (2,)),
            ("forecast_period", (2,)),
            ("model_level_number", (4,)),
            ("level_height", (1,)),
            ("sigma", (1,)),
            ("grid_latitude", (16,)),
            ("grid_longitude", (16,)),
        ]:
            coords = cube.coords(name)
            self.assertEqual(
                len(coords),
                1,
                "expected one {!r} coord, found {}".format(name, len(coords)),
            )
            (coord,) = coords
            self.assertEqual(
                coord.shape,
                shape,
                "coord {!r} shape is {} instead of {!r}.".format(
                    name, coord.shape, shape
                ),
            )

    def test_cube_data(self):
        # Check just a few points of the data.
        cube = self.test_cube
        self.assertArrayAllClose(
            cube.data[:, ::2, 6, 13],
            np.array([[4.218922, 10.074577], [4.626897, 6.520156]]),
            atol=1.0e-6,
        )

    def test_cube_mask(self):
        # Check the data mask : should be just the centre 6x2 section.
        cube = self.test_cube
        mask = np.zeros((2, 4, 16, 16), dtype=bool)
        mask[:, :, 7:9, 5:11] = True
        self.assertArrayEqual(cube.data.mask, mask)


@tests.skip_data
class TestSkipField(tests.IrisTest):
    def test_missing_lbrel(self):
        infile = tests.get_data_path(("FF", "lbrel_missing"))
        with mock.patch("warnings.warn") as warn_fn:
            fields = iris.load(infile)
        self.assertIn(
            "Input field skipped as PPField creation failed : "
            "error = 'Unsupported header release number: -32768'",
            warn_fn.call_args[0][0],
        )
        self.assertEqual(len(fields), 2)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test the cube concatenate mechanism.

"""

# import iris tests first so that some things can be initialised
# before importing anything else.
import iris.tests as tests

import numpy as np
import numpy.ma as ma

import iris.cube
from iris.coords import DimCoord, AuxCoord, CellMeasure, AncillaryVariable
import iris.tests.stock as stock


def _make_cube(
    x, y, data, aux=None, cell_measure=None, ancil=None, offset=0, scalar=None
):
    """
    A convenience test function that creates a custom 2D cube.

    Args:

    * x:
        A (start, stop, step) tuple for specifying the
        x-axis dimensional coordinate points. Bounds are
        automatically guessed.

    * y:
        A (start, stop, step) tuple for specifying the
        y-axis dimensional coordinate points. Bounds are
        automatically guessed.

    * data:
        The data payload for the cube.

    Kwargs:

    * aux:
        A CSV string specifying which points only auxiliary
        coordinates to create. Accepts either of 'x', 'y', 'xy'.

    * offset:
        Offset value to be added to the 'xy' auxiliary coordinate
        points.

    * scalar:
        Create a 'height' scalar coordinate with the given value.

    Returns:
        The newly created 2D :class:`iris.cube.Cube`.

    """
    x_range = np.arange(*x, dtype=np.float32)
    y_range = np.arange(*y, dtype=np.float32)
    x_size = len(x_range)
    y_size = len(y_range)

    cube_data = np.empty((y_size, x_size), dtype=np.float32)
    cube_data[:] = data
    cube = iris.cube.Cube(cube_data)
    coord = DimCoord(y_range, long_name="y", units="1")
    coord.guess_bounds()
    cube.add_dim_coord(coord, 0)
    coord = DimCoord(x_range, long_name="x", units="1")
    coord.guess_bounds()
    cube.add_dim_coord(coord, 1)

    if aux is not None:
        aux = aux.split(",")
        if "y" in aux:
            coord = AuxCoord(y_range * 10, long_name="y-aux", units="1")
            cube.add_aux_coord(coord, (0,))
        if "x" in aux:
            coord = AuxCoord(x_range * 10, long_name="x-aux", units="1")
            cube.add_aux_coord(coord, (1,))
        if "xy" in aux:
            payload = np.arange(y_size * x_size, dtype=np.float32).reshape(
                y_size, x_size
            )
            coord = AuxCoord(
                payload * 100 + offset, long_name="xy-aux", units="1"
            )
            cube.add_aux_coord(coord, (0, 1))

    if cell_measure is not None:
        cell_measure = cell_measure.split(",")
        if "y" in cell_measure:
            cm = CellMeasure(y_range * 10, long_name="y-aux", units="1")
            cube.add_cell_measure(cm, (0,))
        if "x" in cell_measure:
            cm = CellMeasure(x_range * 10, long_name="x-aux", units="1")
            cube.add_cell_measure(cm, (1,))
        if "xy" in cell_measure:
            payload = x_range + y_range[:, np.newaxis]
            cm = CellMeasure(
                payload * 100 + offset, long_name="xy-aux", units="1"
            )
            cube.add_cell_measure(cm, (0, 1))

    if ancil is not None:
        ancil = ancil.split(",")
        if "y" in ancil:
            av = AncillaryVariable(y_range * 10, long_name="y-aux", units="1")
            cube.add_ancillary_variable(av, (0,))
        if "x" in ancil:
            av = AncillaryVariable(x_range * 10, long_name="x-aux", units="1")
            cube.add_ancillary_variable(av, (1,))
        if "xy" in ancil:
            payload = x_range + y_range[:, np.newaxis]
            av = AncillaryVariable(
                payload * 100 + offset, long_name="xy-aux", units="1"
            )
            cube.add_ancillary_variable(av, (0, 1))

    if scalar is not None:
        data = np.array([scalar], dtype=np.float32)
        coord = AuxCoord(data, long_name="height", units="m")
        cube.add_aux_coord(coord, ())

    return cube


def _make_cube_3d(x, y, z, data, aux=None, offset=0):
    """
    A convenience test function that creates a custom 3D cube.

    Args:

    * x:
        A (start, stop, step) tuple for specifying the
        x-axis dimensional coordinate points. Bounds are
        automatically guessed.

    * y:
        A (start, stop, step) tuple for specifying the
        y-axis dimensional coordinate points. Bounds are
        automatically guessed.

    * z:
        A (start, stop, step) tuple for specifying the
        z-axis dimensional coordinate points. Bounds are
        automatically guessed.

    * data:
        The data payload for the cube.

    Kwargs:

    * aux:
        A CSV string specifying which points only auxiliary
        coordinates to create. Accepts either of 'x', 'y', 'z',
        'xy', 'xz', 'yz', 'xyz'.

    * offset:
        Offset value to be added to non-1D auxiliary coordinate
        points.

    Returns:
        The newly created 3D :class:`iris.cube.Cube`.

    """
    x_range = np.arange(*x, dtype=np.float32)
    y_range = np.arange(*y, dtype=np.float32)
    z_range = np.arange(*z, dtype=np.float32)
    x_size, y_size, z_size = len(x_range), len(y_range), len(z_range)

    cube_data = np.empty((x_size, y_size, z_size), dtype=np.float32)
    cube_data[:] = data
    cube = iris.cube.Cube(cube_data)
    coord = DimCoord(z_range, long_name="z", units="1")
    coord.guess_bounds()
    cube.add_dim_coord(coord, 0)
    coord = DimCoord(y_range, long_name="y", units="1")
    coord.guess_bounds()
    cube.add_dim_coord(coord, 1)
    coord = DimCoord(x_range, long_name="x", units="1")
    coord.guess_bounds()
    cube.add_dim_coord(coord, 2)

    if aux is not None:
        aux = aux.split(",")
        if "z" in aux:
            coord = AuxCoord(z_range * 10, long_name="z-aux", units="1")
            cube.add_aux_coord(coord, (0,))
        if "y" in aux:
            coord = AuxCoord(y_range * 10, long_name="y-aux", units="1")
            cube.add_aux_coord(coord, (1,))
        if "x" in aux:
            coord = AuxCoord(x_range * 10, long_name="x-aux", units="1")
            cube.add_aux_coord(coord, (2,))
        if "xy" in aux:
            payload = np.arange(x_size * y_size, dtype=np.float32).reshape(
                y_size, x_size
            )
            coord = AuxCoord(payload + offset, long_name="xy-aux", units="1")
            cube.add_aux_coord(coord, (1, 2))
        if "xz" in aux:
            payload = np.arange(x_size * z_size, dtype=np.float32).reshape(
                z_size, x_size
            )
            coord = AuxCoord(
                payload * 10 + offset, long_name="xz-aux", units="1"
            )
            cube.add_aux_coord(coord, (0, 2))
        if "yz" in aux:
            payload = np.arange(y_size * z_size, dtype=np.float32).reshape(
                z_size, y_size
            )
            coord = AuxCoord(
                payload * 100 + offset, long_name="yz-aux", units="1"
            )
            cube.add_aux_coord(coord, (0, 1))
        if "xyz" in aux:
            payload = np.arange(
                x_size * y_size * z_size, dtype=np.float32
            ).reshape(z_size, y_size, x_size)
            coord = AuxCoord(
                payload * 1000 + offset, long_name="xyz-aux", units="1"
            )
            cube.add_aux_coord(coord, (0, 1, 2))

    return cube


def concatenate(cubes, order=None):
    """
    Explicitly force the contiguous major order of cube data
    alignment to ensure consistent CML crc32 checksums.

    Defaults to contiguous 'C' row-major order.

    """
    if order is None:
        order = "C"

    cubelist = iris.cube.CubeList(cubes)
    result = cubelist.concatenate()

    for cube in result:
        if ma.isMaskedArray(cube.data):
            # cube.data = ma.copy(cube.data, order=order)
            data = np.array(cube.data.data, copy=True, order=order)
            mask = np.array(cube.data.mask, copy=True, order=order)
            fill_value = cube.data.fill_value
            cube.data = ma.array(data, mask=mask, fill_value=fill_value)
        else:
            # cube.data = np.copy(cube.data, order=order)
            cube.data = np.array(cube.data, copy=True, order=order)

    return result


class TestSimple(tests.IrisTest):
    def test_empty(self):
        cubes = iris.cube.CubeList()
        self.assertEqual(concatenate(cubes), iris.cube.CubeList())

    def test_single(self):
        cubes = [stock.simple_2d()]
        self.assertEqual(concatenate(cubes), cubes)

    def test_multi_equal(self):
        cubes = [stock.simple_2d()] * 2
        self.assertEqual(concatenate(cubes), cubes)


class TestNoConcat(tests.IrisTest):
    def test_one_cube_has_anon_dim(self):
        cubes = []
        y = (0, 2)
        cubes.append(_make_cube((0, 2), y, 1))
        cubes.append(_make_cube((2, 4), y, 2))
        cube = _make_cube((4, 6), y, 3)
        cube.remove_coord("x")
        cubes.append(cube)
        result = concatenate(cubes)
        self.assertCML(result, ("concatenate", "concat_anonymous.cml"))
        self.assertEqual(len(result), 2)
        self.assertEqual(result[0].shape, (2, 4))
        self.assertEqual(result[1].shape, (2, 2))

    def test_points_overlap_increasing(self):
        cubes = []
        y = (0, 2)
        cubes.append(_make_cube((0, 2), y, 1))
        cubes.append(_make_cube((1, 3), y, 2))
        result = concatenate(cubes)
        self.assertEqual(len(result), 2)

    def test_points_overlap_decreasing(self):
        cubes = []
        x = (0, 2)
        cubes.append(_make_cube(x, (3, 0, -1), 1))
        cubes.append(_make_cube(x, (1, -1, -1), 2))
        result = concatenate(cubes)
        self.assertEqual(len(result), 2)

    def test_bounds_overlap_increasing(self):
        cubes = []
        y = (0, 2)
        cubes.append(_make_cube((0, 2), y, 1))
        cube = _make_cube((2, 4), y, 1)
        cube.coord("x").bounds = np.array([[0.5, 2.5], [2.5, 3.5]])
        cubes.append(cube)
        result = concatenate(cubes)
        self.assertEqual(len(result), 2)

    def test_bounds_overlap_decreasing(self):
        cubes = []
        y = (0, 2)
        cubes.append(_make_cube((3, 1, -1), y, 1))
        cube = _make_cube((1, -1, -1), y, 2)
        cube.coord("x").bounds = np.array([[2.5, 0.5], [0.5, -0.5]])
        cubes.append(cube)
        result = concatenate(cubes)
        self.assertEqual(len(result), 2)

    def test_scalar_difference(self):
        cubes = []
        y = (0, 2)
        cubes.append(_make_cube((0, 2), y, 1, scalar=10))
        cubes.append(_make_cube((2, 4), y, 2, scalar=20))
        result = concatenate(cubes)
        self.assertEqual(len(result), 2)

    def test_uncommon(self):
        cubes = []
        cubes.append(_make_cube((0, 2), (0, 2), 1))
        cubes.append(_make_cube((2, 4), (2, 4), 2))
        result = concatenate(cubes)
        self.assertEqual(len(result), 2)

    def test_order_difference(self):
        cubes = []
        y = (0, 2)
        cubes.append(_make_cube((0, 2), y, 1))
        cubes.append(_make_cube((6, 1, -1), y, 2))
        result = concatenate(cubes)
        self.assertEqual(len(result), 2)

    def test_cell_measure_missing(self):
        cubes = []
        y = (0, 2)
        cubes.append(_make_cube((0, 2), y, 1, cell_measure="x"))
        cubes.append(_make_cube((2, 4), y, 2))
        result = concatenate(cubes)
        self.assertEqual(len(result), 2)

    def test_ancil_missing(self):
        cubes = []
        y = (0, 2)
        cubes.append(_make_cube((0, 2), y, 1, ancil="x"))
        cubes.append(_make_cube((2, 4), y, 2))
        result = concatenate(cubes)
        self.assertEqual(len(result), 2)


class Test2D(tests.IrisTest):
    def test_masked_and_unmasked(self):
        cubes = []
        y = (0, 2)
        cube = _make_cube((2, 4), y, 2)
        cube.data = ma.asarray(cube.data)
        cubes.append(cube)
        cubes.append(_make_cube((0, 2), y, 1))
        result = concatenate(cubes)
        self.assertEqual(len(result), 1)

    def test_unmasked_and_masked(self):
        cubes = []
        y = (0, 2)
        cubes.append(_make_cube((0, 2), y, 1))
        cube = _make_cube((2, 4), y, 2)
        cube.data = ma.asarray(cube.data)
        cubes.append(cube)
        result = concatenate(cubes)
        self.assertEqual(len(result), 1)

    def test_masked_fill_value(self):
        cubes = []
        y = (0, 2)
        cube = _make_cube((0, 2), y, 1)
        cube.data = ma.asarray(cube.data)
        cube.data.fill_value = 10
        cubes.append(cube)
        cube = _make_cube((2, 4), y, 1)
        cube.data = ma.asarray(cube.data)
        cube.data.fill_value = 20
        cubes.append(cube)
        result = concatenate(cubes)
        self.assertEqual(len(result), 1)

    def test_concat_masked_2x2d(self):
        cubes = []
        y = (0, 2)
        cube = _make_cube((0, 2), y, 1)
        cube.data = ma.asarray(cube.data)
        cube.data[(0, 1), (0, 1)] = ma.masked
        cubes.append(cube)
        cube = _make_cube((2, 4), y, 2)
        cube.data = ma.asarray(cube.data)
        cube.data[(0, 1), (1, 0)] = ma.masked
        cubes.append(cube)
        result = concatenate(cubes)
        self.assertCML(result, ("concatenate", "concat_masked_2x2d.cml"))
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].shape, (2, 4))
        mask = np.array(
            [[True, False, False, True], [False, True, True, False]],
            dtype=np.bool,
        )
        self.assertArrayEqual(result[0].data.mask, mask)

    def test_concat_masked_2y2d(self):
        cubes = []
        x = (0, 2)
        cube = _make_cube(x, (0, 2), 1)
        cube.data = np.ma.asarray(cube.data)
        cube.data[(0, 1), (0, 1)] = ma.masked
        cubes.append(cube)
        cube = _make_cube(x, (2, 4), 2)
        cube.data = ma.asarray(cube.data)
        cube.data[(0, 1), (1, 0)] = ma.masked
        cubes.append(cube)
        result = concatenate(cubes)
        self.assertCML(result, ("concatenate", "concat_masked_2y2d.cml"))
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].shape, (4, 2))
        mask = np.array(
            [[True, False], [False, True], [False, True], [True, False]],
            dtype=np.bool,
        )
        self.assertArrayEqual(result[0].data.mask, mask)

    def test_concat_masked_2y2d_with_concrete_and_lazy(self):
        cubes = []
        x = (0, 2)
        cube = _make_cube(x, (0, 2), 1)
        cube.data = np.ma.asarray(cube.data)
        cube.data[(0, 1), (0, 1)] = ma.masked
        cubes.append(cube)
        cube = _make_cube(x, (2, 4), 2)
        cube.data = ma.asarray(cube.data)
        cube.data[(0, 1), (1, 0)] = ma.masked
        cube.data = cube.lazy_data()
        cubes.append(cube)
        result = concatenate(cubes)
        self.assertCML(result, ("concatenate", "concat_masked_2y2d.cml"))
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].shape, (4, 2))
        mask = np.array(
            [[True, False], [False, True], [False, True], [True, False]],
            dtype=np.bool,
        )
        self.assertArrayEqual(result[0].data.mask, mask)

    def test_concat_masked_2y2d_with_lazy_and_concrete(self):
        cubes = []
        x = (0, 2)
        cube = _make_cube(x, (0, 2), 1)
        cube.data = np.ma.asarray(cube.data)
        cube.data[(0, 1), (0, 1)] = ma.masked
        cube.data = cube.lazy_data()
        cubes.append(cube)
        cube = _make_cube(x, (2, 4), 2)
        cube.data = ma.asarray(cube.data)
        cube.data[(0, 1), (1, 0)] = ma.masked
        cubes.append(cube)
        result = concatenate(cubes)
        self.assertCML(result, ("concatenate", "concat_masked_2y2d.cml"))
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].shape, (4, 2))
        mask = np.array(
            [[True, False], [False, True], [False, True], [True, False]],
            dtype=np.bool,
        )
        self.assertArrayEqual(result[0].data.mask, mask)

    def test_concat_2x2d(self):
        cubes = []
        y = (0, 2)
        cubes.append(_make_cube((0, 4), y, 1))
        cubes.append(_make_cube((4, 6), y, 2))
        result = concatenate(cubes)
        self.assertCML(result, ("concatenate", "concat_2x2d.cml"))
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].shape, (2, 6))

    def test_concat_2y2d(self):
        cubes = []
        x = (0, 2)
        cubes.append(_make_cube(x, (0, 4), 1))
        cubes.append(_make_cube(x, (4, 6), 2))
        result = concatenate(cubes)
        self.assertCML(result, ("concatenate", "concat_2y2d.cml"))
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].shape, (6, 2))

    def test_concat_2x2d_aux_x(self):
        cubes = []
        y = (0, 2)
        cubes.append(_make_cube((0, 4), y, 1, aux="x"))
        cubes.append(_make_cube((4, 6), y, 2, aux="x"))
        result = concatenate(cubes)
        self.assertCML(result, ("concatenate", "concat_2x2d_aux_x.cml"))
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].shape, (2, 6))

    def test_concat_2y2d_aux_x(self):
        cubes = []
        x = (0, 2)
        cubes.append(_make_cube(x, (0, 4), 1, aux="x"))
        cubes.append(_make_cube(x, (4, 6), 2, aux="x"))
        result = concatenate(cubes)
        self.assertCML(result, ("concatenate", "concat_2y2d_aux_x.cml"))
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].shape, (6, 2))

    def test_concat_2x2d_aux_y(self):
        cubes = []
        y = (0, 2)
        cubes.append(_make_cube((0, 4), y, 1, aux="y"))
        cubes.append(_make_cube((4, 6), y, 2, aux="y"))
        result = concatenate(cubes)
        self.assertCML(result, ("concatenate", "concat_2x2d_aux_y.cml"))
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].shape, (2, 6))

    def test_concat_2y2d_aux_y(self):
        cubes = []
        x = (0, 2)
        cubes.append(_make_cube(x, (0, 4), 1, aux="y"))
        cubes.append(_make_cube(x, (4, 6), 2, aux="y"))
        result = concatenate(cubes)
        self.assertCML(result, ("concatenate", "concat_2y2d_aux_y.cml"))
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].shape, (6, 2))

    def test_concat_2x2d_aux_x_y(self):
        cubes = []
        y = (0, 2)
        cubes.append(_make_cube((0, 4), y, 1, aux="x,y"))
        cubes.append(_make_cube((4, 6), y, 2, aux="x,y"))
        result = concatenate(cubes)
        self.assertCML(result, ("concatenate", "concat_2x2d_aux_x_y.cml"))
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].shape, (2, 6))

    def test_concat_2y2d_aux_x_y(self):
        cubes = []
        x = (0, 2)
        cubes.append(_make_cube(x, (0, 4), 1, aux="x,y"))
        cubes.append(_make_cube(x, (4, 6), 2, aux="x,y"))
        result = concatenate(cubes)
        self.assertCML(result, ("concatenate", "concat_2y2d_aux_x_y.cml"))
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].shape, (6, 2))

    def test_concat_2x2d_aux_xy(self):
        cubes = []
        y = (0, 2)
        cubes.append(_make_cube((0, 4), y, 1, aux="xy"))
        cubes.append(_make_cube((4, 6), y, 2, aux="xy"))
        result = concatenate(cubes)
        self.assertCML(result, ("concatenate", "concat_2x2d_aux_xy.cml"))
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].shape, (2, 6))

    def test_concat_2y2d_aux_xy(self):
        cubes = []
        x = (0, 2)
        cubes.append(_make_cube(x, (0, 4), 1, aux="xy"))
        cubes.append(_make_cube(x, (4, 6), 2, aux="xy"))
        result = concatenate(cubes)
        self.assertCML(result, ("concatenate", "concat_2y2d_aux_xy.cml"))
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].shape, (6, 2))

    def test_concat_2x2d_aux_x_xy(self):
        cubes = []
        y = (0, 2)
        cubes.append(_make_cube((0, 4), y, 1, aux="x,xy"))
        cubes.append(_make_cube((4, 6), y, 2, aux="x,xy"))
        result = concatenate(cubes)
        self.assertCML(result, ("concatenate", "concat_2x2d_aux_x_xy.cml"))
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].shape, (2, 6))

    def test_concat_2y2d_aux_x_xy(self):
        cubes = []
        x = (0, 2)
        cubes.append(_make_cube(x, (0, 4), 1, aux="x,xy"))
        cubes.append(_make_cube(x, (4, 6), 2, aux="x,xy"))
        result = concatenate(cubes)
        self.assertCML(result, ("concatenate", "concat_2y2d_aux_x_xy.cml"))
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].shape, (6, 2))

    def test_concat_2x2d_aux_y_xy(self):
        cubes = []
        y = (0, 2)
        cubes.append(_make_cube((0, 4), y, 1, aux="y,xy"))
        cubes.append(_make_cube((4, 6), y, 2, aux="y,xy"))
        result = concatenate(cubes)
        self.assertCML(result, ("concatenate", "concat_2x2d_aux_y_xy.cml"))
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].shape, (2, 6))

    def test_concat_2y2d_aux_y_xy(self):
        cubes = []
        x = (0, 2)
        cubes.append(_make_cube(x, (0, 4), 1, aux="y,xy"))
        cubes.append(_make_cube(x, (4, 6), 2, aux="y,xy"))
        result = concatenate(cubes)
        self.assertCML(result, ("concatenate", "concat_2y2d_aux_y_xy.cml"))
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].shape, (6, 2))

    def test_concat_2x2d_aux_x_y_xy(self):
        cubes = []
        y = (0, 2)
        cubes.append(_make_cube((0, 4), y, 1, aux="x,y,xy"))
        cubes.append(_make_cube((4, 6), y, 2, aux="x,y,xy"))
        result = concatenate(cubes)
        self.assertCML(result, ("concatenate", "concat_2x2d_aux_x_y_xy.cml"))
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].shape, (2, 6))

    def test_concat_2y2d_aux_x_y_xy(self):
        cubes = []
        x = (0, 2)
        cubes.append(_make_cube(x, (0, 4), 1, aux="x,y,xy"))
        cubes.append(_make_cube(x, (4, 6), 2, aux="x,y,xy"))
        result = concatenate(cubes)
        self.assertCML(result, ("concatenate", "concat_2y2d_aux_x_y_xy.cml"))
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].shape, (6, 2))

    def test_concat_2x2d_aux_x_bounds(self):
        cubes = []
        y = (0, 2)
        cube = _make_cube((0, 4), y, 1, aux="x")
        cube.coord("x-aux").guess_bounds()
        cubes.append(cube)
        cube = _make_cube((4, 6), y, 2, aux="x")
        cube.coord("x-aux").guess_bounds()
        cubes.append(cube)
        result = concatenate(cubes)
        self.assertCML(result, ("concatenate", "concat_2x2d_aux_x_bounds.cml"))
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].shape, (2, 6))

    def test_concat_2x2d_aux_xy_bounds(self):
        cubes = []
        y = (0, 2)
        cube = _make_cube((0, 2), y, 1, aux="xy", offset=1)
        coord = cube.coord("xy-aux")
        coord.bounds = np.array(
            [
                1,
                2,
                3,
                4,
                101,
                102,
                103,
                104,
                201,
                202,
                203,
                204,
                301,
                302,
                303,
                304,
            ]
        ).reshape(2, 2, 4)
        cubes.append(cube)
        cube = _make_cube((2, 4), y, 2, aux="xy", offset=2)
        coord = cube.coord("xy-aux")
        coord.bounds = np.array(
            [
                2,
                3,
                4,
                5,
                102,
                103,
                104,
                105,
                202,
                203,
                204,
                205,
                302,
                303,
                304,
                305,
            ]
        ).reshape(2, 2, 4)
        cubes.append(cube)
        result = concatenate(cubes)
        self.assertCML(
            result, ("concatenate", "concat_2x2d_aux_xy_bounds.cml")
        )
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].shape, (2, 4))

    def test_concat_2y2d_cell_measure_x_y_xy(self):
        cubes = []
        x = (0, 2)
        cubes.append(_make_cube(x, (0, 4), 1, cell_measure="x,y,xy"))
        cubes.append(_make_cube(x, (4, 6), 1, cell_measure="x,y,xy"))
        result = concatenate(cubes)
        com = _make_cube(x, (0, 6), 1, cell_measure="x,y,xy")
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].shape, (6, 2))
        self.assertEqual(result[0], com)

    def test_concat_2y2d_ancil_x_y_xy(self):
        cubes = []
        x = (0, 2)
        cubes.append(_make_cube(x, (0, 4), 1, ancil="x,y,xy"))
        cubes.append(_make_cube(x, (4, 6), 1, ancil="x,y,xy"))
        result = concatenate(cubes)
        com = _make_cube(x, (0, 6), 1, ancil="x,y,xy")
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].shape, (6, 2))
        self.assertEqual(result[0], com)


class TestMulti2D(tests.IrisTest):
    def test_concat_4x2d_aux_xy(self):
        cubes = []
        y = (0, 2)
        cubes.append(_make_cube((0, 2), y, 1, aux="xy", offset=1))
        cubes.append(_make_cube((2, 4), y, 2, aux="xy", offset=2))
        y = (2, 4)
        cubes.append(_make_cube((0, 2), y, 3, aux="xy", offset=3))
        cubes.append(_make_cube((2, 4), y, 4, aux="xy", offset=4))
        result = concatenate(cubes)
        self.assertCML(result, ("concatenate", "concat_4x2d_aux_xy.cml"))
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].shape, (4, 4))

    def test_concat_4y2d_aux_xy(self):
        cubes = []
        x = (0, 2)
        cubes.append(_make_cube(x, (0, 2), 1, aux="xy", offset=1))
        cubes.append(_make_cube(x, (2, 4), 2, aux="xy", offset=2))
        x = (2, 4)
        cubes.append(_make_cube(x, (0, 2), 3, aux="xy", offset=3))
        cubes.append(_make_cube(x, (2, 4), 4, aux="xy", offset=4))
        result = concatenate(cubes)
        self.assertCML(result, ("concatenate", "concat_4y2d_aux_xy.cml"))
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].shape, (4, 4))

    def test_concat_4mix2d_aux_xy(self):
        cubes = []
        cubes.append(_make_cube((0, 2), (0, 2), 1, aux="xy", offset=1))
        cubes.append(_make_cube((2, 4), (2, 4), 2, aux="xy", offset=2))
        cubes.append(_make_cube((2, 4), (0, 2), 3, aux="xy", offset=3))
        cubes.append(_make_cube((0, 2), (2, 4), 4, aux="xy", offset=4))
        result = concatenate(cubes)
        self.assertCML(result, ("concatenate", "concat_4mix2d_aux_xy.cml"))
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].shape, (4, 4))

    def test_concat_9x2d_aux_xy(self):
        cubes = []
        y = (0, 2)
        cubes.append(_make_cube((0, 2), y, 1, aux="xy", offset=1))
        cubes.append(_make_cube((2, 4), y, 2, aux="xy", offset=2))
        cubes.append(_make_cube((4, 6), y, 3, aux="xy", offset=3))
        y = (2, 4)
        cubes.append(_make_cube((0, 2), y, 4, aux="xy", offset=4))
        cubes.append(_make_cube((2, 4), y, 5, aux="xy", offset=5))
        cubes.append(_make_cube((4, 6), y, 6, aux="xy", offset=6))
        y = (4, 6)
        cubes.append(_make_cube((0, 2), y, 7, aux="xy", offset=7))
        cubes.append(_make_cube((2, 4), y, 8, aux="xy", offset=8))
        cubes.append(_make_cube((4, 6), y, 9, aux="xy", offset=9))
        result = concatenate(cubes)
        self.assertCML(result, ("concatenate", "concat_9x2d_aux_xy.cml"))
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].shape, (6, 6))

    def test_concat_9y2d_aux_xy(self):
        cubes = []
        x = (0, 2)
        cubes.append(_make_cube(x, (0, 2), 1, aux="xy", offset=1))
        cubes.append(_make_cube(x, (2, 4), 2, aux="xy", offset=2))
        cubes.append(_make_cube(x, (4, 6), 3, aux="xy", offset=3))
        x = (2, 4)
        cubes.append(_make_cube(x, (0, 2), 4, aux="xy", offset=4))
        cubes.append(_make_cube(x, (2, 4), 5, aux="xy", offset=5))
        cubes.append(_make_cube(x, (4, 6), 6, aux="xy", offset=6))
        x = (4, 6)
        cubes.append(_make_cube(x, (0, 2), 7, aux="xy", offset=7))
        cubes.append(_make_cube(x, (2, 4), 8, aux="xy", offset=8))
        cubes.append(_make_cube(x, (4, 6), 9, aux="xy", offset=9))
        result = concatenate(cubes)
        self.assertCML(result, ("concatenate", "concat_9y2d_aux_xy.cml"))
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].shape, (6, 6))

    def test_concat_9mix2d_aux_xy(self):
        cubes = []
        cubes.append(_make_cube((0, 2), (0, 2), 1, aux="xy", offset=1))
        cubes.append(_make_cube((2, 4), (2, 4), 2, aux="xy", offset=2))
        cubes.append(_make_cube((4, 6), (4, 6), 3, aux="xy", offset=3))
        cubes.append(_make_cube((4, 6), (0, 2), 4, aux="xy", offset=4))
        cubes.append(_make_cube((0, 2), (4, 6), 5, aux="xy", offset=5))
        cubes.append(_make_cube((0, 2), (2, 4), 6, aux="xy", offset=6))
        cubes.append(_make_cube((2, 4), (0, 2), 7, aux="xy", offset=7))
        cubes.append(_make_cube((4, 6), (2, 4), 8, aux="xy", offset=8))
        cubes.append(_make_cube((2, 4), (4, 6), 9, aux="xy", offset=9))
        result = concatenate(cubes)
        self.assertCML(result, ("concatenate", "concat_9mix2d_aux_xy.cml"))
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].shape, (6, 6))


class TestMulti2DScalar(tests.IrisTest):
    def test_concat_scalar_4x2d_aux_xy(self):
        cubes = iris.cube.CubeList()
        # Level 1.
        y = (0, 2)
        cubes.append(_make_cube((0, 2), y, 1, aux="xy", offset=1, scalar=10))
        cubes.append(_make_cube((2, 4), y, 2, aux="xy", offset=2, scalar=10))
        y = (2, 4)
        cubes.append(_make_cube((0, 2), y, 3, aux="xy", offset=3, scalar=10))
        cubes.append(_make_cube((2, 4), y, 4, aux="xy", offset=4, scalar=10))
        # Level 2.
        y = (0, 2)
        cubes.append(_make_cube((0, 2), y, 5, aux="xy", offset=1, scalar=20))
        cubes.append(_make_cube((2, 4), y, 6, aux="xy", offset=2, scalar=20))
        y = (2, 4)
        cubes.append(_make_cube((0, 2), y, 7, aux="xy", offset=3, scalar=20))
        cubes.append(_make_cube((2, 4), y, 8, aux="xy", offset=4, scalar=20))
        result = concatenate(cubes)
        self.assertCML(
            result, ("concatenate", "concat_scalar_4x2d_aux_xy.cml")
        )
        self.assertEqual(len(result), 2)
        for cube in result:
            self.assertEqual(cube.shape, (4, 4))

        merged = result.merge()
        self.assertCML(
            merged, ("concatenate", "concat_merged_scalar_4x2d_aux_xy.cml")
        )
        self.assertEqual(len(merged), 1)
        self.assertEqual(merged[0].shape, (2, 4, 4))

        # Test concatenate and merge are commutative operations.
        merged = cubes.merge()
        self.assertCML(
            merged, ("concatenate", "concat_pre_merged_scalar_4x2_aux_xy.cml")
        )
        self.assertEqual(len(merged), 4)

        result = concatenate(merged)
        self.assertCML(
            result, ("concatenate", "concat_merged_scalar_4x2d_aux_xy.cml")
        )
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].shape, (2, 4, 4))

    def test_concat_scalar_4y2d_aux_xy(self):
        cubes = iris.cube.CubeList()
        # Level 1.
        x = (0, 2)
        cubes.append(_make_cube(x, (0, 2), 1, aux="xy", offset=1, scalar=10))
        cubes.append(_make_cube(x, (2, 4), 2, aux="xy", offset=2, scalar=10))
        x = (2, 4)
        cubes.append(_make_cube(x, (0, 2), 3, aux="xy", offset=3, scalar=10))
        cubes.append(_make_cube(x, (2, 4), 4, aux="xy", offset=4, scalar=10))
        # Level 2.
        x = (0, 2)
        cubes.append(_make_cube(x, (0, 2), 5, aux="xy", offset=1, scalar=20))
        cubes.append(_make_cube(x, (2, 4), 6, aux="xy", offset=2, scalar=20))
        x = (2, 4)
        cubes.append(_make_cube(x, (0, 2), 7, aux="xy", offset=3, scalar=20))
        cubes.append(_make_cube(x, (2, 4), 8, aux="xy", offset=4, scalar=20))
        result = concatenate(cubes)
        self.assertCML(
            result, ("concatenate", "concat_scalar_4y2d_aux_xy.cml")
        )
        self.assertEqual(len(result), 2)
        for cube in result:
            self.assertEqual(cube.shape, (4, 4))

        merged = result.merge()
        self.assertCML(
            merged, ("concatenate", "concat_merged_scalar_4y2d_aux_xy.cml")
        )
        self.assertEqual(len(merged), 1)
        self.assertEqual(merged[0].shape, (2, 4, 4))

        # Test concatenate and merge are commutative operations.
        merged = cubes.merge()
        self.assertCML(
            merged, ("concatenate", "concat_pre_merged_scalar_4y2d_aux_xy.cml")
        )
        self.assertEqual(len(merged), 4)

        result = concatenate(merged)
        self.assertEqual(len(result), 1)
        self.assertCML(
            result, ("concatenate", "concat_merged_scalar_4y2d_aux_xy.cml")
        )
        self.assertEqual(result[0].shape, (2, 4, 4))

    def test_concat_scalar_4mix2d_aux_xy(self):
        cubes = iris.cube.CubeList()
        cubes.append(
            _make_cube((0, 2), (0, 2), 1, aux="xy", offset=1, scalar=10)
        )
        cubes.append(
            _make_cube((2, 4), (2, 4), 8, aux="xy", offset=4, scalar=20)
        )
        cubes.append(
            _make_cube((0, 2), (0, 2), 5, aux="xy", offset=1, scalar=20)
        )
        cubes.append(
            _make_cube((2, 4), (0, 2), 2, aux="xy", offset=2, scalar=10)
        )
        cubes.append(
            _make_cube((0, 2), (2, 4), 7, aux="xy", offset=3, scalar=20)
        )
        cubes.append(
            _make_cube((0, 2), (2, 4), 3, aux="xy", offset=3, scalar=10)
        )
        cubes.append(
            _make_cube((2, 4), (2, 4), 4, aux="xy", offset=4, scalar=10)
        )
        cubes.append(
            _make_cube((2, 4), (0, 2), 6, aux="xy", offset=2, scalar=20)
        )
        result = concatenate(cubes)
        self.assertCML(
            result, ("concatenate", "concat_scalar_4mix2d_aux_xy.cml")
        )
        self.assertEqual(len(result), 2)
        for cube in result:
            self.assertEqual(cube.shape, (4, 4))

        merged = result.merge()
        self.assertCML(
            merged, ("concatenate", "concat_merged_scalar_4mix2d_aux_xy.cml")
        )
        self.assertEqual(len(merged), 1)
        self.assertEqual(merged[0].shape, (2, 4, 4))

        # Test concatenate and merge are commutative operations.
        merged = cubes.merge()
        self.assertCML(
            merged,
            ("concatenate", "concat_pre_merged_scalar_4mix2d_aux_xy.cml"),
        )
        self.assertEqual(len(merged), 4)

        result = concatenate(merged)
        self.assertCML(
            result, ("concatenate", "concat_merged_scalar_4mix2d_aux_xy.cml")
        )
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].shape, (2, 4, 4))


class Test3D(tests.IrisTest):
    def _make_group(self, xoff=0, yoff=0, zoff=0, doff=0):
        xoff *= 4
        yoff *= 4
        zoff *= 4
        doff *= 8
        cubes = []
        cubes.append(
            _make_cube_3d(
                (0 + xoff, 2 + xoff),
                (0 + yoff, 2 + yoff),
                (0 + zoff, 2 + zoff),
                1 + doff,
                aux="x,y,z,xy,xz,yz,xyz",
            )
        )
        cubes.append(
            _make_cube_3d(
                (2 + xoff, 4 + xoff),
                (0 + yoff, 2 + yoff),
                (0 + zoff, 2 + zoff),
                2 + doff,
                aux="x,y,z,xy,xz,yz,xyz",
            )
        )
        cubes.append(
            _make_cube_3d(
                (0 + xoff, 2 + xoff),
                (2 + yoff, 4 + yoff),
                (0 + zoff, 2 + zoff),
                3 + doff,
                aux="x,y,z,xy,xz,yz,xyz",
            )
        )
        cubes.append(
            _make_cube_3d(
                (2 + xoff, 4 + xoff),
                (2 + yoff, 4 + yoff),
                (0 + zoff, 2 + zoff),
                4 + doff,
                aux="x,y,z,xy,xz,yz,xyz",
            )
        )

        cubes.append(
            _make_cube_3d(
                (0 + xoff, 2 + xoff),
                (0 + yoff, 2 + yoff),
                (2 + zoff, 4 + zoff),
                5 + doff,
                aux="x,y,z,xy,xz,yz,xyz",
            )
        )
        cubes.append(
            _make_cube_3d(
                (2 + xoff, 4 + xoff),
                (0 + yoff, 2 + yoff),
                (2 + zoff, 4 + zoff),
                6 + doff,
                aux="x,y,z,xy,xz,yz,xyz",
            )
        )
        cubes.append(
            _make_cube_3d(
                (0 + xoff, 2 + xoff),
                (2 + yoff, 4 + yoff),
                (2 + zoff, 4 + zoff),
                7 + doff,
                aux="x,y,z,xy,xz,yz,xyz",
            )
        )
        cubes.append(
            _make_cube_3d(
                (2 + xoff, 4 + xoff),
                (2 + yoff, 4 + yoff),
                (2 + zoff, 4 + zoff),
                8 + doff,
                aux="x,y,z,xy,xz,yz,xyz",
            )
        )

        return cubes

    def test_concat_3d_simple(self):
        cubes = self._make_group()
        result = concatenate(cubes)
        self.assertCML(result, ("concatenate", "concat_3d_simple.cml"))
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].shape, (4, 4, 4))

    def test_concat_3d_mega(self):
        cubes = []
        cubes.extend(self._make_group(xoff=0, doff=0))
        cubes.extend(self._make_group(xoff=1, doff=1))
        cubes.extend(self._make_group(yoff=1, doff=2))
        cubes.extend(self._make_group(xoff=1, yoff=1, doff=3))

        cubes.extend(self._make_group(xoff=0, zoff=1, doff=4))
        cubes.extend(self._make_group(xoff=1, zoff=1, doff=5))
        cubes.extend(self._make_group(yoff=1, zoff=1, doff=6))
        cubes.extend(self._make_group(xoff=1, yoff=1, zoff=1, doff=7))
        result = concatenate(cubes)

        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].shape, (8, 8, 8))


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for the
`iris.aux_factory.OceanSg2Factory` class.

"""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from unittest import mock

from cf_units import Unit
import numpy as np

from iris.aux_factory import OceanSg2Factory
from iris.coords import AuxCoord, DimCoord


class Test___init__(tests.IrisTest):
    def setUp(self):
        self.s = mock.Mock(units=Unit("1"), nbounds=0)
        self.c = mock.Mock(units=Unit("1"), nbounds=0, shape=(1,))
        self.eta = mock.Mock(units=Unit("m"), nbounds=0)
        self.depth = mock.Mock(units=Unit("m"), nbounds=0)
        self.depth_c = mock.Mock(units=Unit("m"), nbounds=0, shape=(1,))
        self.kwargs = dict(
            s=self.s,
            c=self.c,
            eta=self.eta,
            depth=self.depth,
            depth_c=self.depth_c,
        )

    def test_insufficient_coordinates(self):
        with self.assertRaises(ValueError):
            OceanSg2Factory()
        with self.assertRaises(ValueError):
            OceanSg2Factory(
                s=None,
                c=self.c,
                eta=self.eta,
                depth=self.depth,
                depth_c=self.depth_c,
            )
        with self.assertRaises(ValueError):
            OceanSg2Factory(
                s=self.s,
                c=None,
                eta=self.eta,
                depth=self.depth,
                depth_c=self.depth_c,
            )
        with self.assertRaises(ValueError):
            OceanSg2Factory(
                s=self.s,
                c=self.c,
                eta=None,
                depth=self.depth,
                depth_c=self.depth_c,
            )
        with self.assertRaises(ValueError):
            OceanSg2Factory(
                s=self.s,
                c=self.c,
                eta=self.eta,
                depth=None,
                depth_c=self.depth_c,
            )
        with self.assertRaises(ValueError):
            OceanSg2Factory(
                s=self.s,
                c=self.c,
                eta=self.eta,
                depth=self.depth,
                depth_c=None,
            )

    def test_s_too_many_bounds(self):
        self.s.nbounds = 4
        with self.assertRaises(ValueError):
            OceanSg2Factory(**self.kwargs)

    def test_c_too_many_bounds(self):
        self.c.nbounds = 4
        with self.assertRaises(ValueError):
            OceanSg2Factory(**self.kwargs)

    def test_depth_c_non_scalar(self):
        self.depth_c.shape = (2,)
        with self.assertRaises(ValueError):
            OceanSg2Factory(**self.kwargs)

    def test_s_incompatible_units(self):
        self.s.units = Unit("km")
        with self.assertRaises(ValueError):
            OceanSg2Factory(**self.kwargs)

    def test_c_incompatible_units(self):
        self.c.units = Unit("km")
        with self.assertRaises(ValueError):
            OceanSg2Factory(**self.kwargs)

    def test_eta_incompatible_units(self):
        self.eta.units = Unit("km")
        with self.assertRaises(ValueError):
            OceanSg2Factory(**self.kwargs)

    def test_depth_c_incompatible_units(self):
        self.depth_c.units = Unit("km")
        with self.assertRaises(ValueError):
            OceanSg2Factory(**self.kwargs)

    def test_depth_incompatible_units(self):
        self.depth.units = Unit("km")
        with self.assertRaises(ValueError):
            OceanSg2Factory(**self.kwargs)

    def test_promote_c_and_s_units_unknown_to_dimensionless(self):
        c = mock.Mock(units=Unit("unknown"), nbounds=0)
        s = mock.Mock(units=Unit("unknown"), nbounds=0)
        self.kwargs["c"] = c
        self.kwargs["s"] = s
        factory = OceanSg2Factory(**self.kwargs)
        self.assertEqual("1", factory.dependencies["c"].units)
        self.assertEqual("1", factory.dependencies["s"].units)


class Test_dependencies(tests.IrisTest):
    def setUp(self):
        self.s = mock.Mock(units=Unit("1"), nbounds=0)
        self.c = mock.Mock(units=Unit("1"), nbounds=0, shape=(1,))
        self.eta = mock.Mock(units=Unit("m"), nbounds=0)
        self.depth = mock.Mock(units=Unit("m"), nbounds=0)
        self.depth_c = mock.Mock(units=Unit("m"), nbounds=0, shape=(1,))
        self.kwargs = dict(
            s=self.s,
            c=self.c,
            eta=self.eta,
            depth=self.depth,
            depth_c=self.depth_c,
        )

    def test_values(self):
        factory = OceanSg2Factory(**self.kwargs)
        self.assertEqual(factory.dependencies, self.kwargs)


class Test_make_coord(tests.IrisTest):
    @staticmethod
    def coord_dims(coord):
        mapping = dict(s=(0,), c=(0,), eta=(1, 2), depth=(1, 2), depth_c=())
        return mapping[coord.name()]

    @staticmethod
    def derive(s, c, eta, depth, depth_c, coord=True):
        S = (depth_c * s + depth * c) / (depth_c + depth)
        result = eta + (eta + depth) * S
        if coord:
            name = "sea_surface_height_above_reference_ellipsoid"
            result = AuxCoord(
                result,
                standard_name=name,
                units="m",
                attributes=dict(positive="up"),
            )
        return result

    def setUp(self):
        self.s = DimCoord(
            np.linspace(-0.985, -0.014, 36), units="1", long_name="s"
        )
        self.c = DimCoord(
            np.linspace(-0.959, -0.001, 36), units="1", long_name="c"
        )
        self.eta = AuxCoord(
            np.arange(-1, 3, dtype=np.float).reshape(2, 2),
            long_name="eta",
            units="m",
        )
        self.depth = AuxCoord(
            np.array([[5, 200], [1000, 4000]], dtype=np.float),
            long_name="depth",
            units="m",
        )
        self.depth_c = AuxCoord([1], long_name="depth_c", units="m")
        self.kwargs = dict(
            s=self.s,
            c=self.c,
            eta=self.eta,
            depth=self.depth,
            depth_c=self.depth_c,
        )

    def test_derived_points(self):
        # Broadcast expected points given the known dimensional mapping.
        s = self.s.points[..., np.newaxis, np.newaxis]
        c = self.c.points[..., np.newaxis, np.newaxis]
        eta = self.eta.points[np.newaxis, ...]
        depth = self.depth.points[np.newaxis, ...]
        depth_c = self.depth_c.points
        # Calculate the expected result.
        expected_coord = self.derive(s, c, eta, depth, depth_c)
        # Calculate the actual result.
        factory = OceanSg2Factory(**self.kwargs)
        coord = factory.make_coord(self.coord_dims)
        self.assertEqual(expected_coord, coord)


class Test_update(tests.IrisTest):
    def setUp(self):
        self.s = mock.Mock(units=Unit("1"), nbounds=0)
        self.c = mock.Mock(units=Unit("1"), nbounds=0, shape=(1,))
        self.eta = mock.Mock(units=Unit("m"), nbounds=0)
        self.depth = mock.Mock(units=Unit("m"), nbounds=0)
        self.depth_c = mock.Mock(units=Unit("m"), nbounds=0, shape=(1,))
        self.kwargs = dict(
            s=self.s,
            c=self.c,
            eta=self.eta,
            depth=self.depth,
            depth_c=self.depth_c,
        )
        self.factory = OceanSg2Factory(**self.kwargs)

    def test_s(self):
        new_s = mock.Mock(units=Unit("1"), nbounds=0)
        self.factory.update(self.s, new_s)
        self.assertIs(self.factory.s, new_s)

    def test_c(self):
        new_c = mock.Mock(units=Unit("1"), nbounds=0)
        self.factory.update(self.c, new_c)
        self.assertIs(self.factory.c, new_c)

    def test_s_too_many_bounds(self):
        new_s = mock.Mock(units=Unit("1"), nbounds=4)
        with self.assertRaises(ValueError):
            self.factory.update(self.s, new_s)

    def test_c_too_many_bounds(self):
        new_c = mock.Mock(units=Unit("1"), nbounds=4)
        with self.assertRaises(ValueError):
            self.factory.update(self.c, new_c)

    def test_s_incompatible_units(self):
        new_s = mock.Mock(units=Unit("Pa"), nbounds=0)
        with self.assertRaises(ValueError):
            self.factory.update(self.s, new_s)

    def test_c_incompatible_units(self):
        new_c = mock.Mock(units=Unit("Pa"), nbounds=0)
        with self.assertRaises(ValueError):
            self.factory.update(self.c, new_c)

    def test_eta(self):
        new_eta = mock.Mock(units=Unit("m"), nbounds=0)
        self.factory.update(self.eta, new_eta)
        self.assertIs(self.factory.eta, new_eta)

    def test_eta_incompatible_units(self):
        new_eta = mock.Mock(units=Unit("Pa"), nbounds=0)
        with self.assertRaises(ValueError):
            self.factory.update(self.eta, new_eta)

    def test_depth(self):
        new_depth = mock.Mock(units=Unit("m"), nbounds=0)
        self.factory.update(self.depth, new_depth)
        self.assertIs(self.factory.depth, new_depth)

    def test_depth_incompatible_units(self):
        new_depth = mock.Mock(units=Unit("Pa"), nbounds=0)
        with self.assertRaises(ValueError):
            self.factory.update(self.depth, new_depth)

    def test_depth_c(self):
        new_depth_c = mock.Mock(units=Unit("m"), nbounds=0, shape=(1,))
        self.factory.update(self.depth_c, new_depth_c)
        self.assertIs(self.factory.depth_c, new_depth_c)

    def test_depth_c_non_scalar(self):
        new_depth_c = mock.Mock(units=Unit("m"), nbounds=0, shape=(10,))
        with self.assertRaises(ValueError):
            self.factory.update(self.depth_c, new_depth_c)

    def test_depth_c_incompatible_units(self):
        new_depth_c = mock.Mock(units=Unit("Pa"), nbounds=0, shape=(1,))
        with self.assertRaises(ValueError):
            self.factory.update(self.depth_c, new_depth_c)


if __name__ == "__main__":
    tests.main()



"""
Hovmoller Diagram of Monthly Surface Temperature
================================================

This example demonstrates the creation of a Hovmoller diagram with fine control
over plot ticks and labels. The data comes from the Met Office OSTIA project
and has been pre-processed to calculate the monthly mean sea surface
temperature.

"""

import matplotlib.dates as mdates
import matplotlib.pyplot as plt

import iris
import iris.plot as iplt
import iris.quickplot as qplt


def main():
    # load a single cube of surface temperature between +/- 5 latitude
    fname = iris.sample_data_path("ostia_monthly.nc")
    cube = iris.load_cube(
        fname,
        iris.Constraint("surface_temperature", latitude=lambda v: -5 < v < 5),
    )

    # Take the mean over latitude
    cube = cube.collapsed("latitude", iris.analysis.MEAN)

    # Now that we have our data in a nice way, lets create the plot
    # contour with 20 levels
    qplt.contourf(cube, 20)

    # Put a custom label on the y axis
    plt.ylabel("Time / years")

    # Stop matplotlib providing clever axes range padding
    plt.axis("tight")

    # As we are plotting annual variability, put years as the y ticks
    plt.gca().yaxis.set_major_locator(mdates.YearLocator())

    # And format the ticks to just show the year
    plt.gca().yaxis.set_major_formatter(mdates.DateFormatter("%Y"))

    iplt.show()


if __name__ == "__main__":
    main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Integration tests for :mod:`iris.analysis.trajectory`."""

# import iris tests first so that some things can be initialised before
# importing anything else
import iris.tests as tests

import numpy as np

import iris
import iris.tests.stock as istk
from iris._lazy_data import as_lazy_data

from iris.analysis.trajectory import (
    Trajectory,
    interpolate as traj_interpolate,
)


@tests.skip_data
class TestColpex(tests.IrisTest):
    def setUp(self):
        # Load the COLPEX data => TZYX
        path = tests.get_data_path(
            ["PP", "COLPEX", "theta_and_orog_subset.pp"]
        )
        cube = iris.load_cube(path, "air_potential_temperature")
        cube.coord("grid_latitude").bounds = None
        cube.coord("grid_longitude").bounds = None
        # TODO: Workaround until regrid can handle factories
        cube.remove_aux_factory(cube.aux_factories[0])
        cube.remove_coord("surface_altitude")
        self.cube = cube

    def test_trajectory_extraction(self):
        # Pull out a single point - no interpolation required
        single_point = traj_interpolate(
            self.cube,
            [("grid_latitude", [-0.1188]), ("grid_longitude", [359.57958984])],
        )
        expected = self.cube[..., 10, 0].data
        self.assertArrayAllClose(
            single_point[..., 0].data, expected, rtol=2.0e-7
        )
        self.assertCML(
            single_point, ("trajectory", "single_point.cml"), checksum=False
        )

    def test_trajectory_extraction_calc(self):
        # Pull out another point and test against a manually calculated result.
        single_point = [
            ["grid_latitude", [-0.1188]],
            ["grid_longitude", [359.584090412]],
        ]
        scube = self.cube[0, 0, 10:11, 4:6]
        x0 = scube.coord("grid_longitude")[0].points
        x1 = scube.coord("grid_longitude")[1].points
        y0 = scube.data[0, 0]
        y1 = scube.data[0, 1]
        expected = y0 + ((y1 - y0) * ((359.584090412 - x0) / (x1 - x0)))
        trajectory_cube = traj_interpolate(scube, single_point)
        self.assertArrayAllClose(trajectory_cube.data, expected, rtol=2.0e-7)

    def _traj_to_sample_points(self, trajectory):
        sample_points = []
        src_points = trajectory.sampled_points
        for name in src_points[0].keys():
            values = [point[name] for point in src_points]
            sample_points.append((name, values))
        return sample_points

    def test_trajectory_extraction_axis_aligned(self):
        # Extract a simple, axis-aligned trajectory that is similar to an
        # indexing operation.
        # (It's not exactly the same because the source cube doesn't have
        # regular spacing.)
        waypoints = [
            {"grid_latitude": -0.1188, "grid_longitude": 359.57958984},
            {"grid_latitude": -0.1188, "grid_longitude": 359.66870117},
        ]
        trajectory = Trajectory(waypoints, sample_count=100)
        sample_points = self._traj_to_sample_points(trajectory)
        trajectory_cube = traj_interpolate(self.cube, sample_points)
        self.assertCML(
            trajectory_cube, ("trajectory", "constant_latitude.cml")
        )

    def test_trajectory_extraction_zigzag(self):
        # Extract a zig-zag trajectory
        waypoints = [
            {"grid_latitude": -0.1188, "grid_longitude": 359.5886},
            {"grid_latitude": -0.0828, "grid_longitude": 359.6606},
            {"grid_latitude": -0.0468, "grid_longitude": 359.6246},
        ]
        trajectory = Trajectory(waypoints, sample_count=20)
        sample_points = self._traj_to_sample_points(trajectory)
        trajectory_cube = traj_interpolate(self.cube[0, 0], sample_points)
        expected = np.array(
            [
                287.95953369,
                287.9190979,
                287.95550537,
                287.93240356,
                287.83850098,
                287.87869263,
                287.90942383,
                287.9463501,
                287.74365234,
                287.68856812,
                287.75588989,
                287.54611206,
                287.48522949,
                287.53356934,
                287.60217285,
                287.43795776,
                287.59701538,
                287.52468872,
                287.45025635,
                287.52716064,
            ],
            dtype=np.float32,
        )

        self.assertCML(
            trajectory_cube, ("trajectory", "zigzag.cml"), checksum=False
        )
        self.assertArrayAllClose(trajectory_cube.data, expected, rtol=2.0e-7)

    def test_colpex__nearest(self):
        # Check a smallish nearest-neighbour interpolation against a result
        # snapshot.
        test_cube = self.cube[0][0]
        # Test points on a regular grid, a bit larger than the source region.
        xmin, xmax = [
            fn(test_cube.coord(axis="x").points) for fn in (np.min, np.max)
        ]
        ymin, ymax = [
            fn(test_cube.coord(axis="x").points) for fn in (np.min, np.max)
        ]
        fractions = [-0.23, -0.01, 0.27, 0.624, 0.983, 1.052, 1.43]
        x_points = [xmin + frac * (xmax - xmin) for frac in fractions]
        y_points = [ymin + frac * (ymax - ymin) for frac in fractions]
        x_points, y_points = np.meshgrid(x_points, y_points)
        sample_points = [
            ("grid_longitude", x_points.flatten()),
            ("grid_latitude", y_points.flatten()),
        ]
        result = traj_interpolate(test_cube, sample_points, method="nearest")
        expected = [
            288.07168579,
            288.07168579,
            287.9367981,
            287.82736206,
            287.78564453,
            287.8374939,
            287.8374939,
            288.07168579,
            288.07168579,
            287.9367981,
            287.82736206,
            287.78564453,
            287.8374939,
            287.8374939,
            288.07168579,
            288.07168579,
            287.9367981,
            287.82736206,
            287.78564453,
            287.8374939,
            287.8374939,
            288.07168579,
            288.07168579,
            287.9367981,
            287.82736206,
            287.78564453,
            287.8374939,
            287.8374939,
            288.07168579,
            288.07168579,
            287.9367981,
            287.82736206,
            287.78564453,
            287.8374939,
            287.8374939,
            288.07168579,
            288.07168579,
            287.9367981,
            287.82736206,
            287.78564453,
            287.8374939,
            287.8374939,
            288.07168579,
            288.07168579,
            287.9367981,
            287.82736206,
            287.78564453,
            287.8374939,
            287.8374939,
        ]
        self.assertArrayAllClose(result.data, expected)


@tests.skip_data
class TestTriPolar(tests.IrisTest):
    def setUp(self):
        # load data
        cubes = iris.load(
            tests.get_data_path(["NetCDF", "ORCA2", "votemper.nc"])
        )
        cube = cubes[0]
        # The netCDF file has different data types for the points and
        # bounds of 'depth'. This wasn't previously supported, so we
        # emulate that old behaviour.
        b32 = cube.coord("depth").bounds.astype(np.float32)
        cube.coord("depth").bounds = b32
        self.cube = cube
        # define a latitude trajectory (put coords in a different order
        # to the cube, just to be awkward)
        latitudes = list(range(-90, 90, 2))
        longitudes = [-90] * len(latitudes)
        self.sample_points = [
            ("longitude", longitudes),
            ("latitude", latitudes),
        ]

    def test_tri_polar(self):
        # extract
        sampled_cube = traj_interpolate(self.cube, self.sample_points)
        self.assertCML(
            sampled_cube, ("trajectory", "tri_polar_latitude_slice.cml")
        )

    def test_tri_polar_method_linear_fails(self):
        # Try to request linear interpolation.
        # Not allowed, as we have multi-dimensional coords.
        self.assertRaises(
            iris.exceptions.CoordinateMultiDimError,
            traj_interpolate,
            self.cube,
            self.sample_points,
            method="linear",
        )

    def test_tri_polar_method_unknown_fails(self):
        # Try to request unknown interpolation.
        self.assertRaises(
            ValueError,
            traj_interpolate,
            self.cube,
            self.sample_points,
            method="linekar",
        )

    def test_tri_polar__nearest(self):
        # Check a smallish nearest-neighbour interpolation against a result
        # snapshot.
        test_cube = self.cube
        # Use just one 2d layer, just to be faster.
        test_cube = test_cube[0][0]
        # Fix the fill value of the data to zero, just so that we get the same
        # result under numpy < 1.11 as with 1.11.
        # NOTE: numpy<1.11 *used* to assign missing data points into an
        # unmasked array as =0.0, now =fill-value.
        # TODO: arguably, we should support masked data properly in the
        # interpolation routine.  In the legacy code, that is unfortunately
        # just not the case.
        test_cube.data.fill_value = 0

        # Test points on a regular global grid, with unrelated steps + offsets
        # and an extended range of longitude values.
        x_points = np.arange(-185.23, +360.0, 73.123)
        y_points = np.arange(-89.12, +90.0, 42.847)
        x_points, y_points = np.meshgrid(x_points, y_points)
        sample_points = [
            ("longitude", x_points.flatten()),
            ("latitude", y_points.flatten()),
        ]
        result = traj_interpolate(test_cube, sample_points, method="nearest")
        expected = [
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            12.13186264,
            10.69991493,
            9.86881161,
            7.08723927,
            9.04308414,
            12.56258678,
            10.63761806,
            9.19426727,
            28.93525505,
            23.85289955,
            26.94649506,
            0.0,
            27.88831711,
            28.65439224,
            23.39414215,
            26.78363228,
            13.53453922,
            0.0,
            17.41485596,
            0.0,
            0.0,
            13.0413475,
            0.0,
            17.10849571,
            -1.67040622,
            -1.64783156,
            0.0,
            -1.97898054,
            -1.67642927,
            -1.65173221,
            -1.623945,
            0.0,
        ]

        self.assertArrayAllClose(result.data, expected)


class TestLazyData(tests.IrisTest):
    def test_hybrid_height(self):
        cube = istk.simple_4d_with_hybrid_height()
        # Put a lazy array into the cube so we can test deferred loading.
        cube.data = as_lazy_data(cube.data)

        traj = (
            ("grid_latitude", [20.5, 21.5, 22.5, 23.5]),
            ("grid_longitude", [31, 32, 33, 34]),
        )
        xsec = traj_interpolate(cube, traj, method="nearest")

        # Check that creating the trajectory hasn't led to the original
        # data being loaded.
        self.assertTrue(cube.has_lazy_data())
        self.assertCML([cube, xsec], ("trajectory", "hybrid_height.cml"))


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for
:func:`iris.fileformats.pp_load_rules._reshape_vector_args`.

"""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import numpy as np

from iris.fileformats.pp_load_rules import _reshape_vector_args


class TestEmpty(tests.IrisTest):
    def test(self):
        result = _reshape_vector_args([])
        self.assertEqual(result, [])


class TestSingleArg(tests.IrisTest):
    def _check(self, result, expected):
        self.assertEqual(len(result), len(expected))
        for result_arr, expected_arr in zip(result, expected):
            self.assertArrayEqual(result_arr, expected_arr)

    def test_nochange(self):
        points = np.array([[1, 2, 3], [4, 5, 6]])
        result = _reshape_vector_args([(points, (0, 1))])
        expected = [points]
        self._check(result, expected)

    def test_bad_dimensions(self):
        points = np.array([[1, 2, 3], [4, 5, 6]])
        with self.assertRaisesRegex(ValueError, "Length"):
            _reshape_vector_args([(points, (0, 1, 2))])

    def test_scalar(self):
        points = 5
        result = _reshape_vector_args([(points, ())])
        expected = [points]
        self._check(result, expected)

    def test_nonarray(self):
        points = [[1, 2, 3], [4, 5, 6]]
        result = _reshape_vector_args([(points, (0, 1))])
        expected = [np.array(points)]
        self._check(result, expected)

    def test_transpose(self):
        points = np.array([[1, 2, 3], [4, 5, 6]])
        result = _reshape_vector_args([(points, (1, 0))])
        expected = [points.T]
        self._check(result, expected)

    def test_extend(self):
        points = np.array([[1, 2, 3, 4], [21, 22, 23, 24], [31, 32, 33, 34]])
        result = _reshape_vector_args([(points, (1, 3))])
        expected = [points.reshape(1, 3, 1, 4)]
        self._check(result, expected)


class TestMultipleArgs(tests.IrisTest):
    def _check(self, result, expected):
        self.assertEqual(len(result), len(expected))
        for result_arr, expected_arr in zip(result, expected):
            self.assertArrayEqual(result_arr, expected_arr)

    def test_nochange(self):
        a1 = np.array([[1, 2, 3], [4, 5, 6]])
        a2 = np.array([[0, 2, 4], [7, 8, 9]])
        result = _reshape_vector_args([(a1, (0, 1)), (a2, (0, 1))])
        expected = [a1, a2]
        self._check(result, expected)

    def test_array_and_scalar(self):
        a1 = [[1, 2, 3], [3, 4, 5]]
        a2 = 5
        result = _reshape_vector_args([(a1, (0, 1)), (a2, ())])
        expected = [a1, np.array([[5]])]
        self._check(result, expected)

    def test_transpose(self):
        a1 = np.array([[1, 2, 3], [4, 5, 6]])
        a2 = np.array([[0, 2, 4], [7, 8, 9]])
        result = _reshape_vector_args([(a1, (0, 1)), (a2, (1, 0))])
        expected = [a1, a2.T]
        self._check(result, expected)

    def test_incompatible(self):
        # Does not enforce compatibility of results.
        a1 = np.array([1, 2])
        a2 = np.array([1, 2, 3])
        result = _reshape_vector_args([(a1, (0,)), (a2, (0,))])
        expected = [a1, a2]
        self._check(result, expected)

    def test_extend(self):
        a1 = np.array([[1, 2, 3], [4, 5, 6]])
        a2 = np.array([11, 12, 13])
        result = _reshape_vector_args([(a1, (0, 1)), (a2, (1,))])
        expected = [a1, a2.reshape(1, 3)]
        self._check(result, expected)

    def test_extend_transpose(self):
        a1 = np.array([[1, 2, 3], [4, 5, 6]])
        a2 = np.array([11, 12, 13])
        result = _reshape_vector_args([(a1, (1, 0)), (a2, (1,))])
        expected = [a1.T, a2.reshape(1, 3)]
        self._check(result, expected)

    def test_double_extend(self):
        a1 = np.array([[1, 2, 3], [4, 5, 6]])
        a2 = np.array(1)
        result = _reshape_vector_args([(a1, (0, 2)), (a2, ())])
        expected = [a1.reshape(2, 1, 3), a2.reshape(1, 1, 1)]
        self._check(result, expected)

    def test_triple(self):
        a1 = np.array([[1, 2, 3, 4]])
        a2 = np.array([3, 4])
        a3 = np.array(7)
        result = _reshape_vector_args([(a1, (0, 2)), (a2, (1,)), (a3, ())])
        expected = [
            a1.reshape(1, 1, 4),
            a2.reshape(1, 2, 1),
            a3.reshape(1, 1, 1),
        ]
        self._check(result, expected)


if __name__ == "__main__":
    tests.main()



"""
Deriving the Coriolis Frequency Over the Globe
==============================================

This code computes the Coriolis frequency and stores it in a cube with
associated metadata. It then plots the Coriolis frequency on an orthographic
projection.

"""

import cartopy.crs as ccrs
import matplotlib.pyplot as plt
import numpy as np


import iris
from iris.coord_systems import GeogCS
import iris.plot as iplt


def main():
    # Start with arrays for latitudes and longitudes, with a given number of
    # coordinates in the arrays.
    coordinate_points = 200
    longitudes = np.linspace(-180.0, 180.0, coordinate_points)
    latitudes = np.linspace(-90.0, 90.0, coordinate_points)
    lon2d, lat2d = np.meshgrid(longitudes, latitudes)

    # Omega is the Earth's rotation rate, expressed in radians per second
    omega = 7.29e-5

    # The data for our cube is the Coriolis frequency,
    # `f = 2 * omega * sin(phi)`, which is computed for each grid point over
    # the globe from the 2-dimensional latitude array.
    data = 2.0 * omega * np.sin(np.deg2rad(lat2d))

    # We now need to define a coordinate system for the plot.
    # Here we'll use GeogCS; 6371229 is the radius of the Earth in metres.
    cs = GeogCS(6371229)

    # The Iris coords module turns the latitude list into a coordinate array.
    # Coords then applies an appropriate standard name and unit to it.
    lat_coord = iris.coords.DimCoord(
        latitudes, standard_name="latitude", units="degrees", coord_system=cs
    )

    # The above process is repeated for the longitude coordinates.
    lon_coord = iris.coords.DimCoord(
        longitudes, standard_name="longitude", units="degrees", coord_system=cs
    )

    # Now we add bounds to our latitude and longitude coordinates.
    # We want simple, contiguous bounds for our regularly-spaced coordinate
    # points so we use the guess_bounds() method of the coordinate. For more
    # complex coordinates, we could derive and set the bounds manually.
    lat_coord.guess_bounds()
    lon_coord.guess_bounds()

    # Now we input our data array into the cube.
    new_cube = iris.cube.Cube(
        data,
        standard_name="coriolis_parameter",
        units="s-1",
        dim_coords_and_dims=[(lat_coord, 0), (lon_coord, 1)],
    )

    # Now let's plot our cube, along with coastlines, a title and an
    # appropriately-labelled colour bar:
    ax = plt.axes(projection=ccrs.Orthographic())
    ax.coastlines(resolution="10m")
    mesh = iplt.pcolormesh(new_cube, cmap="seismic")
    tick_levels = [-0.00012, -0.00006, 0.0, 0.00006, 0.00012]
    plt.colorbar(
        mesh,
        orientation="horizontal",
        label="s-1",
        ticks=tick_levels,
        format="%.1e",
    )
    plt.title("Coriolis frequency")
    plt.show()


if __name__ == "__main__":
    main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Test the function :func:`iris._lazy data.as_lazy_data`."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from unittest import mock

import dask.array as da
import dask.config
import numpy as np
import numpy.ma as ma

from iris._lazy_data import as_lazy_data, _optimum_chunksize


class Test_as_lazy_data(tests.IrisTest):
    def test_lazy(self):
        data = da.from_array(np.arange(24).reshape((2, 3, 4)), chunks="auto")
        result = as_lazy_data(data)
        self.assertIsInstance(result, da.core.Array)

    def test_real(self):
        data = np.arange(24).reshape((2, 3, 4))
        result = as_lazy_data(data)
        self.assertIsInstance(result, da.core.Array)

    def test_masked(self):
        data = np.ma.masked_greater(np.arange(24), 10)
        result = as_lazy_data(data)
        self.assertIsInstance(result, da.core.Array)

    def test_non_default_chunks(self):
        data = np.arange(24)
        chunks = (12,)
        lazy_data = as_lazy_data(data, chunks=chunks)
        (result,) = np.unique(lazy_data.chunks)
        self.assertEqual(result, 24)

    def test_with_masked_constant(self):
        masked_data = ma.masked_array([8], mask=True)
        masked_constant = masked_data[0]
        result = as_lazy_data(masked_constant)
        self.assertIsInstance(result, da.core.Array)


class Test__optimised_chunks(tests.IrisTest):
    # Stable, known chunksize for testing.
    FIXED_CHUNKSIZE_LIMIT = 1024 * 1024 * 64

    @staticmethod
    def _dummydata(shape):
        return mock.Mock(spec=da.core.Array, dtype=np.dtype("f4"), shape=shape)

    def test_chunk_size_limiting(self):
        # Check default chunksizes for large data (with a known size limit).
        given_shapes_and_resulting_chunks = [
            ((16, 1024, 1024), (16, 1024, 1024)),  # largest unmodified
            ((17, 1011, 1022), (8, 1011, 1022)),
            ((16, 1024, 1025), (8, 1024, 1025)),
            ((1, 17, 1011, 1022), (1, 8, 1011, 1022)),
            ((17, 1, 1011, 1022), (8, 1, 1011, 1022)),
            ((11, 2, 1011, 1022), (5, 2, 1011, 1022)),
        ]
        err_fmt = "Result of optimising chunks {} was {}, expected {}"
        for (shape, expected) in given_shapes_and_resulting_chunks:
            chunks = _optimum_chunksize(
                shape, shape, limit=self.FIXED_CHUNKSIZE_LIMIT
            )
            msg = err_fmt.format(shape, chunks, expected)
            self.assertEqual(chunks, expected, msg)

    def test_chunk_size_expanding(self):
        # Check the expansion of small chunks, (with a known size limit).
        given_shapes_and_resulting_chunks = [
            ((1, 100, 100), (16, 100, 100), (16, 100, 100)),
            ((1, 100, 100), (5000, 100, 100), (1667, 100, 100)),
            ((3, 300, 200), (10000, 3000, 2000), (3, 1500, 2000)),
            ((3, 300, 200), (10000, 300, 2000), (27, 300, 2000)),
            ((3, 300, 200), (8, 300, 2000), (8, 300, 2000)),
            ((3, 300, 200), (117, 300, 1000), (39, 300, 1000)),
        ]
        err_fmt = "Result of optimising shape={};chunks={} was {}, expected {}"
        for (shape, fullshape, expected) in given_shapes_and_resulting_chunks:
            chunks = _optimum_chunksize(
                chunks=shape, shape=fullshape, limit=self.FIXED_CHUNKSIZE_LIMIT
            )
            msg = err_fmt.format(fullshape, shape, chunks, expected)
            self.assertEqual(chunks, expected, msg)

    def test_chunk_expanding_equal_division(self):
        # Check that expansion chooses equal chunk sizes as far as possible.

        # Table of test cases:
        #   (input-chunkshape, full-shape, size-limit, result-chunkshape)
        testcases_chunksin_fullshape_limit_result = [
            ((4,), (12,), 15, (12,)),  # gives a single chunk, of size 12
            ((4,), (13,), 15, (8,)),  # chooses chunks of 8+5, better than 12+1
            ((4,), (16,), 15, (8,)),  # 8+8 is better than 12+4; 16 is too big.
            ((4,), (96,), 15, (12,)),  # 12 is largest 'allowed'
            ((4,), (96,), 31, (24,)),  # 28 doesn't divide 96 so neatly,
            # A multi-dimensional case, where trailing dims are 'filled'.
            ((4, 5, 100), (25, 10, 200), 16 * 2000, (16, 10, 200)),
            # Equivalent case with additional initial dimensions.
            (
                (1, 1, 4, 5, 100),
                (3, 5, 25, 10, 200),
                16 * 2000,
                (1, 1, 16, 10, 200),
            ),  # effectively the same as the previous.
        ]
        err_fmt_main = (
            "Main chunks result of optimising "
            "chunks={},shape={},limit={} "
            "was {}, expected {}"
        )
        for (
            chunks,
            shape,
            limit,
            expected_result,
        ) in testcases_chunksin_fullshape_limit_result:
            result = _optimum_chunksize(
                chunks=chunks, shape=shape, limit=limit, dtype=np.dtype("b1")
            )
            msg = err_fmt_main.format(
                chunks, shape, limit, result, expected_result
            )
            self.assertEqual(result, expected_result, msg)

    def test_default_chunksize(self):
        # Check that the "ideal" chunksize is taken from the dask config.
        with dask.config.set({"array.chunk-size": "20b"}):
            chunks = _optimum_chunksize(
                (1, 8), shape=(400, 20), dtype=np.dtype("f4")
            )
            self.assertEqual(chunks, (1, 4))

    def test_default_chunks_limiting(self):
        # Check that chunking is still controlled when no specific 'chunks'
        # is passed.
        limitcall_patch = self.patch("iris._lazy_data._optimum_chunksize")
        test_shape = (3, 2, 4)
        data = self._dummydata(test_shape)
        as_lazy_data(data)
        self.assertEqual(
            limitcall_patch.call_args_list,
            [
                mock.call(
                    list(test_shape), shape=test_shape, dtype=np.dtype("f4")
                )
            ],
        )

    def test_shapeless_data(self):
        # Check that chunk optimisation is skipped if shape contains a zero.
        limitcall_patch = self.patch("iris._lazy_data._optimum_chunksize")
        test_shape = (2, 1, 0, 2)
        data = self._dummydata(test_shape)
        as_lazy_data(data, chunks=test_shape)
        self.assertFalse(limitcall_patch.called)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.

from iris.analysis._interpolation import get_xy_dim_coords, snapshot_grid
import iris.experimental.regrid as eregrid


class AreaWeightedRegridder:
    """
    This class provides support for performing area-weighted regridding.

    """

    def __init__(self, src_grid_cube, target_grid_cube, mdtol=1):
        """
        Create an area-weighted regridder for conversions between the source
        and target grids.

        Args:

        * src_grid_cube:
            The :class:`~iris.cube.Cube` providing the source grid.
        * target_grid_cube:
            The :class:`~iris.cube.Cube` providing the target grid.

        Kwargs:

        * mdtol (float):
            Tolerance of missing data. The value returned in each element of
            the returned array will be masked if the fraction of masked data
            exceeds mdtol. mdtol=0 means no missing data is tolerated while
            mdtol=1 will mean the resulting element will be masked if and only
            if all the contributing elements of data are masked.
            Defaults to 1.

        .. Note::

            Both source and target cubes must have an XY grid defined by
            separate X and Y dimensions with dimension coordinates.
            All of the XY dimension coordinates must also be bounded, and have
            the same cooordinate system.

        """
        # Snapshot the state of the source cube to ensure that the regridder is
        # impervious to external changes to the original cubes.
        self._src_grid = snapshot_grid(src_grid_cube)

        # Missing data tolerance.
        if not (0 <= mdtol <= 1):
            msg = "Value for mdtol must be in range 0 - 1, got {}."
            raise ValueError(msg.format(mdtol))
        self._mdtol = mdtol

        # Store regridding information
        _regrid_info = (
            eregrid._regrid_area_weighted_rectilinear_src_and_grid__prepare(
                src_grid_cube, target_grid_cube
            )
        )
        (
            src_x,
            src_y,
            src_x_dim,
            src_y_dim,
            self.grid_x,
            self.grid_y,
            self.meshgrid_x,
            self.meshgrid_y,
            self.weights_info,
        ) = _regrid_info

    def __call__(self, cube):
        """
        Regrid this :class:`~iris.cube.Cube` onto the target grid of
        this :class:`AreaWeightedRegridder`.

        The given cube must be defined with the same grid as the source
        grid used to create this :class:`AreaWeightedRegridder`.

        If the source cube has lazy data, the returned cube will also
        have lazy data.

        Args:

        * cube:
            A :class:`~iris.cube.Cube` to be regridded.

        Returns:
            A cube defined with the horizontal dimensions of the target
            and the other dimensions from this cube. The data values of
            this cube will be converted to values on the new grid using
            area-weighted regridding.

        .. note::

            If the source cube has lazy data,
            `chunks <https://docs.dask.org/en/latest/array-chunks.html>`__
            in the horizontal dimensions will be combined before regridding.

        """
        src_x, src_y = get_xy_dim_coords(cube)
        if (src_x, src_y) != self._src_grid:
            raise ValueError(
                "The given cube is not defined on the same "
                "source grid as this regridder."
            )
        src_x_dim = cube.coord_dims(src_x)[0]
        src_y_dim = cube.coord_dims(src_y)[0]
        _regrid_info = (
            src_x,
            src_y,
            src_x_dim,
            src_y_dim,
            self.grid_x,
            self.grid_y,
            self.meshgrid_x,
            self.meshgrid_y,
            self.weights_info,
        )
        return eregrid._regrid_area_weighted_rectilinear_src_and_grid__perform(
            cube, _regrid_info, mdtol=self._mdtol
        )



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for :func:`iris.analysis._interpolation.get_xy_dim_coords`.

"""

# import iris tests first so that some things can be initialised
# before importing anything else.
import iris.tests as tests

import copy

import numpy as np

from iris.analysis._interpolation import get_xy_dim_coords
import iris.coords
import iris.coord_systems
import iris.experimental.regrid
import iris.tests.stock


class TestGetXYCoords(tests.IrisTest):
    @tests.skip_data
    def test_grid_lat_lon(self):
        cube = iris.tests.stock.realistic_4d()
        x, y = get_xy_dim_coords(cube)
        self.assertIs(x, cube.coord("grid_longitude"))
        self.assertIs(y, cube.coord("grid_latitude"))

    def test_lat_lon(self):
        cube = iris.tests.stock.lat_lon_cube()
        x, y = get_xy_dim_coords(cube)
        self.assertIs(x, cube.coord("longitude"))
        self.assertIs(y, cube.coord("latitude"))

    def test_projection_coords(self):
        cube = iris.tests.stock.lat_lon_cube()
        cube.coord("longitude").rename("projection_x_coordinate")
        cube.coord("latitude").rename("projection_y_coordinate")
        x, y = get_xy_dim_coords(cube)
        self.assertIs(x, cube.coord("projection_x_coordinate"))
        self.assertIs(y, cube.coord("projection_y_coordinate"))

    @tests.skip_data
    def test_missing_x_coord(self):
        cube = iris.tests.stock.realistic_4d()
        cube.remove_coord("grid_longitude")
        with self.assertRaises(ValueError):
            get_xy_dim_coords(cube)

    @tests.skip_data
    def test_missing_y_coord(self):
        cube = iris.tests.stock.realistic_4d()
        cube.remove_coord("grid_latitude")
        with self.assertRaises(ValueError):
            get_xy_dim_coords(cube)

    @tests.skip_data
    def test_multiple_coords(self):
        cube = iris.tests.stock.realistic_4d()
        cs = iris.coord_systems.GeogCS(6371229)
        time_coord = cube.coord("time")
        time_dims = cube.coord_dims(time_coord)
        lat_coord = iris.coords.DimCoord(
            np.arange(time_coord.shape[0]),
            standard_name="latitude",
            units="degrees",
            coord_system=cs,
        )
        cube.remove_coord(time_coord)
        cube.add_dim_coord(lat_coord, time_dims)
        model_level_coord = cube.coord("model_level_number")
        model_level_dims = cube.coord_dims(model_level_coord)
        lon_coord = iris.coords.DimCoord(
            np.arange(model_level_coord.shape[0]),
            standard_name="longitude",
            units="degrees",
            coord_system=cs,
        )
        cube.remove_coord(model_level_coord)
        cube.add_dim_coord(lon_coord, model_level_dims)

        with self.assertRaises(ValueError):
            get_xy_dim_coords(cube)

        cube.remove_coord("grid_latitude")
        cube.remove_coord("grid_longitude")

        x, y = get_xy_dim_coords(cube)
        self.assertIs(x, lon_coord)
        self.assertIs(y, lat_coord)

    def test_no_coordsystem(self):
        cube = iris.tests.stock.lat_lon_cube()
        for coord in cube.coords():
            coord.coord_system = None
        x, y = get_xy_dim_coords(cube)
        self.assertIs(x, cube.coord("longitude"))
        self.assertIs(y, cube.coord("latitude"))

    def test_one_coordsystem(self):
        cube = iris.tests.stock.lat_lon_cube()
        cube.coord("longitude").coord_system = None
        with self.assertRaises(ValueError):
            get_xy_dim_coords(cube)

    def test_different_coordsystem(self):
        cube = iris.tests.stock.lat_lon_cube()

        lat_cs = copy.copy(cube.coord("latitude").coord_system)
        lat_cs.semi_major_axis = 7000000
        cube.coord("latitude").coord_system = lat_cs

        lon_cs = copy.copy(cube.coord("longitude").coord_system)
        lon_cs.semi_major_axis = 7000001
        cube.coord("longitude").coord_system = lon_cs

        with self.assertRaises(ValueError):
            get_xy_dim_coords(cube)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for the
`iris.aux_factory.OceanSigmaFactory` class.

"""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from unittest import mock

from cf_units import Unit
import numpy as np

from iris.aux_factory import OceanSigmaFactory
from iris.coords import AuxCoord, DimCoord


class Test___init__(tests.IrisTest):
    def setUp(self):
        self.sigma = mock.Mock(units=Unit("1"), nbounds=0)
        self.eta = mock.Mock(units=Unit("m"), nbounds=0)
        self.depth = mock.Mock(units=Unit("m"), nbounds=0)
        self.kwargs = dict(sigma=self.sigma, eta=self.eta, depth=self.depth)

    def test_insufficient_coordinates(self):
        with self.assertRaises(ValueError):
            OceanSigmaFactory()
        with self.assertRaises(ValueError):
            OceanSigmaFactory(sigma=None, eta=self.eta, depth=self.depth)
        with self.assertRaises(ValueError):
            OceanSigmaFactory(sigma=self.sigma, eta=None, depth=self.depth)
        with self.assertRaises(ValueError):
            OceanSigmaFactory(sigma=self.sigma, eta=self.eta, depth=None)

    def test_sigma_too_many_bounds(self):
        self.sigma.nbounds = 4
        with self.assertRaises(ValueError):
            OceanSigmaFactory(**self.kwargs)

    def test_sigma_incompatible_units(self):
        self.sigma.units = Unit("km")
        with self.assertRaises(ValueError):
            OceanSigmaFactory(**self.kwargs)

    def test_eta_incompatible_units(self):
        self.eta.units = Unit("km")
        with self.assertRaises(ValueError):
            OceanSigmaFactory(**self.kwargs)

    def test_depth_incompatible_units(self):
        self.depth.units = Unit("km")
        with self.assertRaises(ValueError):
            OceanSigmaFactory(**self.kwargs)

    def test_promote_sigma_units_unknown_to_dimensionless(self):
        sigma = mock.Mock(units=Unit("unknown"), nbounds=0)
        self.kwargs["sigma"] = sigma
        factory = OceanSigmaFactory(**self.kwargs)
        self.assertEqual("1", factory.dependencies["sigma"].units)


class Test_dependencies(tests.IrisTest):
    def setUp(self):
        self.sigma = mock.Mock(units=Unit("1"), nbounds=0)
        self.eta = mock.Mock(units=Unit("m"), nbounds=0)
        self.depth = mock.Mock(units=Unit("m"), nbounds=0)
        self.kwargs = dict(sigma=self.sigma, eta=self.eta, depth=self.depth)

    def test_values(self):
        factory = OceanSigmaFactory(**self.kwargs)
        self.assertEqual(factory.dependencies, self.kwargs)


class Test_make_coord(tests.IrisTest):
    @staticmethod
    def coord_dims(coord):
        mapping = dict(sigma=(0,), eta=(1, 2), depth=(1, 2))
        return mapping[coord.name()]

    @staticmethod
    def derive(sigma, eta, depth, coord=True):
        result = eta + sigma * (depth + eta)
        if coord:
            name = "sea_surface_height_above_reference_ellipsoid"
            result = AuxCoord(
                result,
                standard_name=name,
                units="m",
                attributes=dict(positive="up"),
            )
        return result

    def setUp(self):
        self.sigma = DimCoord(
            np.linspace(-0.05, -1, 5), long_name="sigma", units="1"
        )
        self.eta = AuxCoord(
            np.arange(-1, 3, dtype=np.float).reshape(2, 2),
            long_name="eta",
            units="m",
        )
        self.depth = AuxCoord(
            np.arange(4, dtype=np.float).reshape(2, 2) * 1e3,
            long_name="depth",
            units="m",
        )
        self.kwargs = dict(sigma=self.sigma, eta=self.eta, depth=self.depth)

    def test_derived_points(self):
        # Broadcast expected points given the known dimensional mapping.
        sigma = self.sigma.points[..., np.newaxis, np.newaxis]
        eta = self.eta.points[np.newaxis, ...]
        depth = self.depth.points[np.newaxis, ...]
        # Calculate the expected result.
        expected_coord = self.derive(sigma, eta, depth)
        # Calculate the actual result.
        factory = OceanSigmaFactory(**self.kwargs)
        coord = factory.make_coord(self.coord_dims)
        self.assertEqual(expected_coord, coord)


class Test_update(tests.IrisTest):
    def setUp(self):
        self.sigma = mock.Mock(units=Unit("1"), nbounds=0)
        self.eta = mock.Mock(units=Unit("m"), nbounds=0)
        self.depth = mock.Mock(units=Unit("m"), nbounds=0)
        self.kwargs = dict(sigma=self.sigma, eta=self.eta, depth=self.depth)
        self.factory = OceanSigmaFactory(**self.kwargs)

    def test_sigma(self):
        new_sigma = mock.Mock(units=Unit("1"), nbounds=0)
        self.factory.update(self.sigma, new_sigma)
        self.assertIs(self.factory.sigma, new_sigma)

    def test_sigma_too_many_bounds(self):
        new_sigma = mock.Mock(units=Unit("1"), nbounds=4)
        with self.assertRaises(ValueError):
            self.factory.update(self.sigma, new_sigma)

    def test_sigma_incompatible_units(self):
        new_sigma = mock.Mock(units=Unit("Pa"), nbounds=0)
        with self.assertRaises(ValueError):
            self.factory.update(self.sigma, new_sigma)

    def test_eta(self):
        new_eta = mock.Mock(units=Unit("m"), nbounds=0)
        self.factory.update(self.eta, new_eta)
        self.assertIs(self.factory.eta, new_eta)

    def test_eta_incompatible_units(self):
        new_eta = mock.Mock(units=Unit("Pa"), nbounds=0)
        with self.assertRaises(ValueError):
            self.factory.update(self.eta, new_eta)

    def test_depth(self):
        new_depth = mock.Mock(units=Unit("m"), nbounds=0)
        self.factory.update(self.depth, new_depth)
        self.assertIs(self.factory.depth, new_depth)

    def test_depth_incompatible_units(self):
        new_depth = mock.Mock(units=Unit("Pa"), nbounds=0)
        with self.assertRaises(ValueError):
            self.factory.update(self.depth, new_depth)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Test function :func:`iris.util.mask_cube"""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import numpy as np
import numpy.ma as ma

from iris.util import mask_cube
from iris.tests.stock import sample_2d_latlons
from iris.tests.stock import make_bounds_discontiguous_at_point


def full2d_global():
    return sample_2d_latlons(transformed=True)


@tests.skip_data
class Test(tests.IrisTest):
    def setUp(self):
        # Set up a 2d cube with a masked discontiguity to test masking
        # of 2-dimensional cubes
        self.cube_2d = full2d_global()
        make_bounds_discontiguous_at_point(self.cube_2d, 3, 3)

    def test_mask_cube_2d(self):
        # This tests the masking of a 2d data array
        cube = self.cube_2d
        discontiguity_array = ma.getmaskarray(cube.data).copy()
        expected = cube.copy()

        # Remove mask so that we can pass an unmasked data set to
        # mask_discontiguities, and check that it masks the correct point by
        # comparing with masked data
        cube.data.mask = ma.nomask
        returned = mask_cube(cube, discontiguity_array)
        self.assertTrue(np.all(expected.data.mask == returned.data.mask))


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.

# import iris tests first so that some things can be initialised before
# importing anything else
import iris.tests as tests

import numpy as np
import numpy.ma as ma
import unittest

import iris
import iris.analysis
import iris.coord_systems
import iris.coords


class TestAggregateBy(tests.IrisTest):
    def setUp(self):
        #
        # common
        #
        cs_latlon = iris.coord_systems.GeogCS(6371229)
        points = np.arange(3, dtype=np.float32) * 3
        coord_lat = iris.coords.DimCoord(
            points, "latitude", units="degrees", coord_system=cs_latlon
        )
        coord_lon = iris.coords.DimCoord(
            points, "longitude", units="degrees", coord_system=cs_latlon
        )

        #
        # single coordinate aggregate-by
        #
        a = np.arange(9, dtype=np.int32).reshape(3, 3) + 1
        b = np.arange(36, dtype=np.int32).reshape(36, 1, 1)
        data = b * a

        self.cube_single = iris.cube.Cube(
            data, long_name="temperature", units="kelvin"
        )

        z_points = np.array(
            [
                1,
                2,
                2,
                3,
                3,
                3,
                4,
                4,
                4,
                4,
                5,
                5,
                5,
                5,
                5,
                6,
                6,
                6,
                6,
                6,
                6,
                7,
                7,
                7,
                7,
                7,
                7,
                7,
                8,
                8,
                8,
                8,
                8,
                8,
                8,
                8,
            ],
            dtype=np.int32,
        )
        self.coord_z_single = iris.coords.AuxCoord(
            z_points, long_name="height", units="m"
        )

        model_level = iris.coords.DimCoord(
            np.arange(z_points.size),
            standard_name="model_level_number",
            units="1",
        )

        self.cube_single.add_aux_coord(self.coord_z_single, 0)
        self.cube_single.add_dim_coord(model_level, 0)
        self.cube_single.add_dim_coord(coord_lon, 1)
        self.cube_single.add_dim_coord(coord_lat, 2)

        #
        # multi coordinate aggregate-by
        #
        a = np.arange(9, dtype=np.int32).reshape(3, 3) + 1
        b = np.arange(20, dtype=np.int32).reshape(20, 1, 1)
        data = b * a

        self.cube_multi = iris.cube.Cube(
            data, long_name="temperature", units="kelvin"
        )

        z1_points = np.array(
            [1, 1, 1, 1, 1, 2, 2, 2, 2, 3, 4, 4, 4, 4, 4, 1, 5, 5, 2, 2],
            dtype=np.int32,
        )
        self.coord_z1_multi = iris.coords.AuxCoord(
            z1_points, long_name="height", units="m"
        )
        z2_points = np.array(
            [1, 1, 1, 3, 3, 3, 5, 5, 5, 7, 7, 7, 9, 9, 9, 11, 11, 11, 3, 3],
            dtype=np.int32,
        )
        self.coord_z2_multi = iris.coords.AuxCoord(
            z2_points, long_name="level", units="1"
        )

        model_level = iris.coords.DimCoord(
            np.arange(z1_points.size),
            standard_name="model_level_number",
            units="1",
        )

        self.cube_multi.add_aux_coord(self.coord_z1_multi, 0)
        self.cube_multi.add_aux_coord(self.coord_z2_multi, 0)
        self.cube_multi.add_dim_coord(model_level, 0)
        self.cube_multi.add_dim_coord(coord_lon.copy(), 1)
        self.cube_multi.add_dim_coord(coord_lat.copy(), 2)

        #
        # expected data results
        #
        self.single_expected = np.array(
            [
                [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]],
                [[1.5, 3.0, 4.5], [6.0, 7.5, 9.0], [10.5, 12.0, 13.5]],
                [[4.0, 8.0, 12.0], [16.0, 20.0, 24.0], [28.0, 32.0, 36.0]],
                [[7.5, 15.0, 22.5], [30.0, 37.5, 45.0], [52.5, 60.0, 67.5]],
                [[12.0, 24.0, 36.0], [48.0, 60.0, 72.0], [84.0, 96.0, 108.0]],
                [
                    [17.5, 35.0, 52.5],
                    [70.0, 87.5, 105.0],
                    [122.5, 140.0, 157.5],
                ],
                [
                    [24.0, 48.0, 72.0],
                    [96.0, 120.0, 144.0],
                    [168.0, 192.0, 216.0],
                ],
                [
                    [31.5, 63.0, 94.5],
                    [126.0, 157.5, 189.0],
                    [220.5, 252.0, 283.5],
                ],
            ],
            dtype=np.float64,
        )

        row1 = [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]]
        row2 = [
            list(np.sqrt([2.5, 10.0, 22.5])),
            list(np.sqrt([40.0, 62.5, 90.0])),
            list(np.sqrt([122.5, 160.0, 202.5])),
        ]
        row3 = [
            list(np.sqrt([16.66666667, 66.66666667, 150.0])),
            list(np.sqrt([266.66666667, 416.66666667, 600.0])),
            list(np.sqrt([816.66666667, 1066.66666667, 1350.0])),
        ]
        row4 = [
            list(np.sqrt([57.5, 230.0, 517.5])),
            list(np.sqrt([920.0, 1437.5, 2070.0])),
            list(np.sqrt([2817.5, 3680.0, 4657.5])),
        ]
        row5 = [
            list(np.sqrt([146.0, 584.0, 1314.0])),
            list(np.sqrt([2336.0, 3650.0, 5256.0])),
            list(np.sqrt([7154.0, 9344.0, 11826.0])),
        ]
        row6 = [
            list(np.sqrt([309.16666667, 1236.66666667, 2782.5])),
            list(np.sqrt([4946.66666667, 7729.16666667, 11130.0])),
            list(np.sqrt([15149.16666667, 19786.66666667, 25042.5])),
        ]
        row7 = [
            list(np.sqrt([580.0, 2320.0, 5220.0])),
            list(np.sqrt([9280.0, 14500.0, 20880.0])),
            list(np.sqrt([28420.0, 37120.0, 46980.0])),
        ]
        row8 = [
            list(np.sqrt([997.5, 3990.0, 8977.5])),
            list(np.sqrt([15960.0, 24937.5, 35910.0])),
            list(np.sqrt([48877.5, 63840.0, 80797.5])),
        ]
        self.single_rms_expected = np.array(
            [row1, row2, row3, row4, row5, row6, row7, row8], dtype=np.float64
        )

        self.multi_expected = np.array(
            [
                [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]],
                [[3.5, 7.0, 10.5], [14.0, 17.5, 21.0], [24.5, 28.0, 31.5]],
                [[14.0, 28.0, 42.0], [56.0, 70.0, 84.0], [98.0, 112.0, 126.0]],
                [[7.0, 14.0, 21.0], [28.0, 35.0, 42.0], [49.0, 56.0, 63.0]],
                [[9.0, 18.0, 27.0], [36.0, 45.0, 54.0], [63.0, 72.0, 81.0]],
                [[10.5, 21.0, 31.5], [42.0, 52.5, 63.0], [73.5, 84.0, 94.5]],
                [[13.0, 26.0, 39.0], [52.0, 65.0, 78.0], [91.0, 104.0, 117.0]],
                [
                    [15.0, 30.0, 45.0],
                    [60.0, 75.0, 90.0],
                    [105.0, 120.0, 135.0],
                ],
                [
                    [16.5, 33.0, 49.5],
                    [66.0, 82.5, 99.0],
                    [115.5, 132.0, 148.5],
                ],
            ],
            dtype=np.float64,
        )

    def test_single(self):
        # mean group-by with single coordinate name.
        aggregateby_cube = self.cube_single.aggregated_by(
            "height", iris.analysis.MEAN
        )
        self.assertCML(
            aggregateby_cube, ("analysis", "aggregated_by", "single.cml")
        )

        # mean group-by with single coordinate.
        aggregateby_cube = self.cube_single.aggregated_by(
            self.coord_z_single, iris.analysis.MEAN
        )
        self.assertCML(
            aggregateby_cube, ("analysis", "aggregated_by", "single.cml")
        )

        np.testing.assert_almost_equal(
            aggregateby_cube.data, self.single_expected
        )

        # rms group-by with single coordinate name.
        aggregateby_cube = self.cube_single.aggregated_by(
            "height", iris.analysis.RMS
        )
        self.assertCML(
            aggregateby_cube, ("analysis", "aggregated_by", "single_rms.cml")
        )

        # rms group-by with single coordinate.
        aggregateby_cube = self.cube_single.aggregated_by(
            self.coord_z_single, iris.analysis.RMS
        )
        self.assertCML(
            aggregateby_cube, ("analysis", "aggregated_by", "single_rms.cml")
        )

        np.testing.assert_almost_equal(
            aggregateby_cube.data, self.single_rms_expected
        )

    def test_single_shared(self):
        z2_points = np.arange(36, dtype=np.int32)
        coord_z2 = iris.coords.AuxCoord(
            z2_points, long_name="wibble", units="1"
        )
        self.cube_single.add_aux_coord(coord_z2, 0)

        # group-by with single coordinate name on shared axis.
        aggregateby_cube = self.cube_single.aggregated_by(
            "height", iris.analysis.MEAN
        )
        self.assertCML(
            aggregateby_cube,
            ("analysis", "aggregated_by", "single_shared.cml"),
        )

        # group-by with single coordinate on shared axis.
        aggregateby_cube = self.cube_single.aggregated_by(
            self.coord_z_single, iris.analysis.MEAN
        )
        self.assertCML(
            aggregateby_cube,
            ("analysis", "aggregated_by", "single_shared.cml"),
        )

        np.testing.assert_almost_equal(
            aggregateby_cube.data, self.single_expected
        )

    def test_single_shared_circular(self):
        points = np.arange(36) * 10.0
        circ_coord = iris.coords.DimCoord(
            points, long_name="circ_height", units="degrees", circular=True
        )
        self.cube_single.add_aux_coord(circ_coord, 0)

        # group-by with single coordinate name on shared axis.
        aggregateby_cube = self.cube_single.aggregated_by(
            "height", iris.analysis.MEAN
        )
        self.assertCML(
            aggregateby_cube,
            ("analysis", "aggregated_by", "single_shared_circular.cml"),
        )

        # group-by with single coordinate on shared axis.
        coord = self.cube_single.coords("height")
        aggregateby_cube = self.cube_single.aggregated_by(
            coord, iris.analysis.MEAN
        )
        self.assertCML(
            aggregateby_cube,
            ("analysis", "aggregated_by", "single_shared_circular.cml"),
        )
        np.testing.assert_almost_equal(
            aggregateby_cube.data, self.single_expected
        )

    def test_multi(self):
        # group-by with multiple coordinate names.
        aggregateby_cube = self.cube_multi.aggregated_by(
            ["height", "level"], iris.analysis.MEAN
        )
        self.assertCML(
            aggregateby_cube, ("analysis", "aggregated_by", "multi.cml")
        )

        # group-by with multiple coordinate names (different order).
        aggregateby_cube = self.cube_multi.aggregated_by(
            ["level", "height"], iris.analysis.MEAN
        )
        self.assertCML(
            aggregateby_cube, ("analysis", "aggregated_by", "multi.cml")
        )

        # group-by with multiple coordinates.
        aggregateby_cube = self.cube_multi.aggregated_by(
            [self.coord_z1_multi, self.coord_z2_multi], iris.analysis.MEAN
        )
        self.assertCML(
            aggregateby_cube, ("analysis", "aggregated_by", "multi.cml")
        )

        # group-by with multiple coordinates (different order).
        aggregateby_cube = self.cube_multi.aggregated_by(
            [self.coord_z2_multi, self.coord_z1_multi], iris.analysis.MEAN
        )
        self.assertCML(
            aggregateby_cube, ("analysis", "aggregated_by", "multi.cml")
        )

        np.testing.assert_almost_equal(
            aggregateby_cube.data, self.multi_expected
        )

    def test_multi_shared(self):
        z3_points = np.arange(20, dtype=np.int32)
        coord_z3 = iris.coords.AuxCoord(
            z3_points, long_name="sigma", units="1"
        )
        z4_points = np.arange(19, -1, -1, dtype=np.int32)
        coord_z4 = iris.coords.AuxCoord(
            z4_points, long_name="gamma", units="1"
        )

        self.cube_multi.add_aux_coord(coord_z3, 0)
        self.cube_multi.add_aux_coord(coord_z4, 0)

        # group-by with multiple coordinate names on shared axis.
        aggregateby_cube = self.cube_multi.aggregated_by(
            ["height", "level"], iris.analysis.MEAN
        )
        self.assertCML(
            aggregateby_cube, ("analysis", "aggregated_by", "multi_shared.cml")
        )

        # group-by with multiple coordinate names on shared axis (different
        # order).
        aggregateby_cube = self.cube_multi.aggregated_by(
            ["level", "height"], iris.analysis.MEAN
        )
        self.assertCML(
            aggregateby_cube, ("analysis", "aggregated_by", "multi_shared.cml")
        )

        # group-by with multiple coordinates on shared axis.
        aggregateby_cube = self.cube_multi.aggregated_by(
            [self.coord_z1_multi, self.coord_z2_multi], iris.analysis.MEAN
        )
        self.assertCML(
            aggregateby_cube, ("analysis", "aggregated_by", "multi_shared.cml")
        )

        # group-by with multiple coordinates on shared axis (different order).
        aggregateby_cube = self.cube_multi.aggregated_by(
            [self.coord_z2_multi, self.coord_z1_multi], iris.analysis.MEAN
        )
        self.assertCML(
            aggregateby_cube, ("analysis", "aggregated_by", "multi_shared.cml")
        )

        np.testing.assert_almost_equal(
            aggregateby_cube.data, self.multi_expected
        )

    def test_easy(self):
        data = np.array(
            [[6, 10, 12, 18], [8, 12, 14, 20], [18, 12, 10, 6]],
            dtype=np.float32,
        )
        cube = iris.cube.Cube(data, long_name="temperature", units="kelvin")

        llcs = iris.coord_systems.GeogCS(6371229)
        cube.add_aux_coord(
            iris.coords.AuxCoord(
                np.array([0, 0, 10], dtype=np.float32),
                "latitude",
                units="degrees",
                coord_system=llcs,
            ),
            0,
        )
        cube.add_aux_coord(
            iris.coords.AuxCoord(
                np.array([0, 0, 10, 10], dtype=np.float32),
                "longitude",
                units="degrees",
                coord_system=llcs,
            ),
            1,
        )

        #
        # Easy mean aggregate test by each coordinate.
        #
        aggregateby_cube = cube.aggregated_by("longitude", iris.analysis.MEAN)
        np.testing.assert_almost_equal(
            aggregateby_cube.data,
            np.array(
                [[8.0, 15.0], [10.0, 17.0], [15.0, 8.0]], dtype=np.float32
            ),
        )
        self.assertCML(
            aggregateby_cube,
            ("analysis", "aggregated_by", "easy.cml"),
            checksum=False,
        )

        aggregateby_cube = cube.aggregated_by("latitude", iris.analysis.MEAN)
        np.testing.assert_almost_equal(
            aggregateby_cube.data,
            np.array(
                [[7.0, 11.0, 13.0, 19.0], [18.0, 12.0, 10.0, 6.0]],
                dtype=np.float32,
            ),
        )

        #
        # Easy max aggregate test by each coordinate.
        #
        aggregateby_cube = cube.aggregated_by("longitude", iris.analysis.MAX)
        np.testing.assert_almost_equal(
            aggregateby_cube.data,
            np.array(
                [[10.0, 18.0], [12.0, 20.0], [18.0, 10.0]], dtype=np.float32
            ),
        )

        aggregateby_cube = cube.aggregated_by("latitude", iris.analysis.MAX)
        np.testing.assert_almost_equal(
            aggregateby_cube.data,
            np.array(
                [[8.0, 12.0, 14.0, 20.0], [18.0, 12.0, 10.0, 6.0]],
                dtype=np.float32,
            ),
        )

        #
        # Easy sum aggregate test by each coordinate.
        #
        aggregateby_cube = cube.aggregated_by("longitude", iris.analysis.SUM)
        np.testing.assert_almost_equal(
            aggregateby_cube.data,
            np.array(
                [[16.0, 30.0], [20.0, 34.0], [30.0, 16.0]], dtype=np.float32
            ),
        )

        aggregateby_cube = cube.aggregated_by("latitude", iris.analysis.SUM)
        np.testing.assert_almost_equal(
            aggregateby_cube.data,
            np.array(
                [[14.0, 22.0, 26.0, 38.0], [18.0, 12.0, 10.0, 6.0]],
                dtype=np.float32,
            ),
        )

        #
        # Easy percentile aggregate test by each coordinate.
        #
        aggregateby_cube = cube.aggregated_by(
            "longitude", iris.analysis.PERCENTILE, percent=25
        )
        np.testing.assert_almost_equal(
            aggregateby_cube.data,
            np.array(
                [[7.0, 13.5], [9.0, 15.5], [13.5, 7.0]], dtype=np.float32
            ),
        )

        aggregateby_cube = cube.aggregated_by(
            "latitude", iris.analysis.PERCENTILE, percent=25
        )
        np.testing.assert_almost_equal(
            aggregateby_cube.data,
            np.array(
                [[6.5, 10.5, 12.5, 18.5], [18.0, 12.0, 10.0, 6.0]],
                dtype=np.float32,
            ),
        )

        #
        # Easy root mean square aggregate test by each coordinate.
        #
        aggregateby_cube = cube.aggregated_by("longitude", iris.analysis.RMS)
        row = [
            list(np.sqrt([68.0, 234.0])),
            list(np.sqrt([104.0, 298.0])),
            list(np.sqrt([234.0, 68.0])),
        ]
        np.testing.assert_almost_equal(
            aggregateby_cube.data, np.array(row, dtype=np.float32)
        )

        aggregateby_cube = cube.aggregated_by("latitude", iris.analysis.RMS)
        row = [
            list(np.sqrt([50.0, 122.0, 170.0, 362.0])),
            [18.0, 12.0, 10.0, 6.0],
        ]
        np.testing.assert_almost_equal(
            aggregateby_cube.data, np.array(row, dtype=np.float32)
        )

    def test_single_missing(self):
        # aggregation correctly handles masked data
        mask = np.vstack(
            (
                np.array([[[0, 1, 0], [1, 0, 1], [0, 1, 0]]]).repeat(
                    26, axis=0
                ),
                np.zeros([10, 3, 3]),
            )
        )
        self.cube_single.data = ma.array(self.cube_single.data, mask=mask)
        single_expected = ma.masked_invalid(
            [
                [
                    [0.0, np.nan, 0.0],
                    [np.nan, 0.0, np.nan],
                    [0.0, np.nan, 0.0],
                ],
                [
                    [1.5, np.nan, 4.5],
                    [np.nan, 7.5, np.nan],
                    [10.5, np.nan, 13.5],
                ],
                [
                    [4.0, np.nan, 12.0],
                    [np.nan, 20.0, np.nan],
                    [28.0, np.nan, 36.0],
                ],
                [
                    [7.5, np.nan, 22.5],
                    [np.nan, 37.5, np.nan],
                    [52.5, np.nan, 67.5],
                ],
                [
                    [12.0, np.nan, 36.0],
                    [np.nan, 60.0, np.nan],
                    [84.0, np.nan, 108.0],
                ],
                [
                    [17.5, np.nan, 52.5],
                    [np.nan, 87.5, np.nan],
                    [122.5, np.nan, 157.5],
                ],
                [
                    [24.0, 53.0, 72.0],
                    [106.0, 120.0, 159.0],
                    [168.0, 212.0, 216.0],
                ],
                [
                    [31.5, 63.0, 94.5],
                    [126.0, 157.5, 189.0],
                    [220.5, 252.0, 283.5],
                ],
            ]
        )
        aggregateby_cube = self.cube_single.aggregated_by(
            "height", iris.analysis.MEAN
        )
        self.assertCML(
            aggregateby_cube,
            ("analysis", "aggregated_by", "single_missing.cml"),
            checksum=False,
        )
        self.assertMaskedArrayAlmostEqual(
            aggregateby_cube.data, single_expected
        )

    def test_multi_missing(self):
        # aggregation correctly handles masked data
        mask = np.vstack(
            (
                np.array([[[0, 1, 0], [1, 0, 1], [0, 1, 0]]]).repeat(
                    16, axis=0
                ),
                np.ones([2, 3, 3]),
                np.zeros([2, 3, 3]),
            )
        )
        self.cube_multi.data = ma.array(self.cube_multi.data, mask=mask)
        multi_expected = ma.masked_invalid(
            [
                [
                    [1.0, np.nan, 3.0],
                    [np.nan, 5.0, np.nan],
                    [7.0, np.nan, 9.0],
                ],
                [
                    [3.5, np.nan, 10.5],
                    [np.nan, 17.5, np.nan],
                    [24.5, np.nan, 31.5],
                ],
                [
                    [14.0, 37.0, 42.0],
                    [74.0, 70.0, 111.0],
                    [98.0, 148.0, 126.0],
                ],
                [
                    [7.0, np.nan, 21.0],
                    [np.nan, 35.0, np.nan],
                    [49.0, np.nan, 63.0],
                ],
                [
                    [9.0, np.nan, 27.0],
                    [np.nan, 45.0, np.nan],
                    [63.0, np.nan, 81.0],
                ],
                [
                    [10.5, np.nan, 31.5],
                    [np.nan, 52.5, np.nan],
                    [73.5, np.nan, 94.5],
                ],
                [
                    [13.0, np.nan, 39.0],
                    [np.nan, 65.0, np.nan],
                    [91.0, np.nan, 117.0],
                ],
                [
                    [15.0, np.nan, 45.0],
                    [np.nan, 75.0, np.nan],
                    [105.0, np.nan, 135.0],
                ],
                [
                    [np.nan, np.nan, np.nan],
                    [np.nan, np.nan, np.nan],
                    [np.nan, np.nan, np.nan],
                ],
            ]
        )
        aggregateby_cube = self.cube_multi.aggregated_by(
            ["height", "level"], iris.analysis.MEAN
        )
        self.assertCML(
            aggregateby_cube,
            ("analysis", "aggregated_by", "multi_missing.cml"),
            checksum=False,
        )
        self.assertMaskedArrayAlmostEqual(
            aggregateby_cube.data, multi_expected
        )

    def test_returned_weights(self):
        self.assertRaises(
            ValueError,
            self.cube_single.aggregated_by,
            "height",
            iris.analysis.MEAN,
            returned=True,
        )
        self.assertRaises(
            ValueError,
            self.cube_single.aggregated_by,
            "height",
            iris.analysis.MEAN,
            weights=[1, 2, 3, 4, 5],
        )


if __name__ == "__main__":
    unittest.main()
